"""This module contains logic to interleave a computation graph (an intervention graph) with the computation graph of a model.

The :class:`InterventionProxy <nnsight.intervention.InterventionProxy>` class extends the functionality of a base nnsight.tracing.Proxy.Proxy object and makes it easier for users to interact with.

:func:`intervene() <nnsight.intervention.InterventionProtocol.intervene>` is the entry hook into the models computation graph in order to interleave an intervention graph.

The :class:`HookModel <nnsight.intervention.HookModel>` provides a context manager for adding input and output hooks to modules and removing them upon context exit.
"""

from __future__ import annotations

import inspect
from collections import defaultdict
from contextlib import AbstractContextManager
from typing import Any, Callable, Collection, Dict, List, Tuple, Union

import torch
from torch.utils.hooks import RemovableHandle
from typing_extensions import Self

from . import util
from .contexts.Conditional import Conditional
from .tracing import protocols
from .tracing.Graph import Graph
from .tracing.Node import Node
from .tracing.protocols import Protocol
from .tracing.Proxy import Proxy


class InterventionProxy(Proxy):
    """Sub-class for Proxy that adds additional user functionality to proxies.

    Examples:

        Saving a proxy so it is not deleted at the completion of it's listeners is enabled with ``.save()``:

        .. code-block:: python

            with model.trace('The Eiffel Tower is in the city of'):
                hidden_states = model.lm_head.input.save()
                logits = model.lm_head.output.save()

            print(hidden_states)
            print(logits)
    """

    def __init__(self, node: Node) -> None:
        super().__init__(node)

        self.__dict__["_grad"] = None

        self._grad: InterventionProxy

    def save(self) -> InterventionProxy:
        """Method when called, indicates to the intervention graph to not delete the tensor values of the result.

        Returns:
            InterventionProxy: Proxy.
        """

        # Add a 'lock' node with the save proxy as an argument to ensure the values are never deleted.
        # This is because 'lock' nodes never actually get set and therefore there will always be a
        # dependency for the save proxy.

        protocols.LockProtocol.add(self.node)

        return self

    def stop(self) -> InterventionProxy:
        """Method when called, indicates to the intervention graph to stop the execution of the model after this Proxy/Node is completed..

        Returns:
            InterventionProxy: Proxy.
        """

        protocols.EarlyStopProtocol.add(self.node.graph, self.node)

        return self

    def update(self, value: Union[Node, Any]) -> InterventionProxy:
        """Updates the value of the Proxy via the creation of the UpdateProtocol node.

        Args:
            - value (Union[Node, Any]): New proxy value.

        Returns:
            InterventionProxy: Proxy.

        .. codeb-block:: python
            with model.trace(input) as tracer:
                num = tracer.apply(int, 0)
                num.update(5)
        """

        return protocols.UpdateProtocol.add(self.node, value)

    @property
    def grad(self) -> InterventionProxy:
        """
        Calling denotes the user wishes to get the grad of proxy tensor and therefore we create a Proxy of that request.
        Only generates a proxy the first time it is references otherwise return the already set one.

        Returns:
            Proxy: Grad proxy.
        """
        if self._grad is None:

            self.__dict__["_grad"] = protocols.GradProtocol.add(self.node)

        return self._grad

    @grad.setter
    def grad(self, value: Union[InterventionProxy, Any]) -> None:
        """
        Calling denotes the user wishes to set the grad of this proxy tensor and therefore we create a Proxy of that request via a SwapProtocol.

        Args:
            value (Union[InterventionProxy, Any]): Value to set output to.
        """
        protocols.SwapProtocol.add(self.grad.node, value)

        self.__dict__["_grad"] = None

    def __call__(self, *args, **kwargs) -> Self:

        # We don't want to call backward on fake tensors.
        # We also want to track the number of times .backward() has been called so .grad on a Proxy refers to the right backward pass.
        if (
            self.node.target is util.fetch_attr
            and isinstance(self.node.args[1], str)
            and self.node.args[1] == "backward"
        ):

            # Clear all .grad proxies so allow users to get the ,.grad of the next backward pass.
            for node in self.node.graph.nodes.values():

                try:

                    if node.proxy._grad is not None:

                        node.proxy.__dict__["_grad"] = None

                except ReferenceError:
                    pass

            # Use GradProtocol to increment the tracking of the number of times .backward() has been called.
            protocols.GradProtocol.increment(self.node.graph)

            return self.node.create(
                proxy_value=None,
                target=Proxy.proxy_call,
                args=[self.node] + list(args),
                kwargs=kwargs,
            )

        return super().__call__(*args, **kwargs)

    def __setattr__(
        self, key: Union[InterventionProxy, Any], value: Union[Self, Any]
    ) -> None:

        # We catch setting .grad as that is a special Protocol vs. setting attributes generally.
        if key == "grad":
            return getattr(self.__class__, key).fset(self, value)

        return super().__setattr__(key, value)

    @property
    def shape(self) -> Collection[torch.Size]:
        """Property to retrieve the shape of the traced proxy value or real value.

        Returns:
            Union[torch.Size,Collection[torch.Size]]: Proxy value shape or collection of shapes.
        """

        if not self.node.attached():

            return util.apply(self.value, lambda x: x.shape, torch.Tensor)

        # If we haven't scanned in a proxy_value, just return a proxy to get the attribute.
        if self.node.proxy_value is inspect._empty:

            return super().__getattr__("shape")

        return util.apply(
            self.node.proxy_value, lambda x: x.shape, torch.Tensor
        )

    @property
    def device(self) -> Collection[torch.device]:
        """Property to retrieve the device of the traced proxy value or real value.

        Returns:
            Union[torch.Size,Collection[torch.device]]: Proxy value device or collection of devices.
        """

        if not self.node.attached():

            return util.apply(self.value, lambda x: x.device, torch.Tensor)

        # If we haven't scanned in a proxy_value, just return a proxy to get the attribute.
        if self.node.proxy_value is inspect._empty:

            return super().__getattr__("device")

        return util.apply(
            self.node.proxy_value, lambda x: x.device, torch.Tensor
        )

    @property
    def dtype(self) -> Collection[torch.device]:
        """Property to retrieve the dtype of the traced proxy value or real value.

        Returns:
            Union[torch.Size,Collection[torch.dtype]]: Proxy value dtype or collection of dtypes.
        """

        if not self.node.attached():

            return util.apply(self.value, lambda x: x.dtype, torch.Tensor)

        # If we haven't scanned in a proxy_value, just return a proxy to get the attribute.
        if self.node.proxy_value is inspect._empty:

            return super().__getattr__("dtype")

        return util.apply(
            self.node.proxy_value, lambda x: x.dtype, torch.Tensor
        )


class InterventionProtocol(Protocol):
    """Primary Protocol that handles tracking and injecting inputs and outputs from a torch model into the overall intervention Graph.
    Uses an attachment on the Graph to store the names of nodes that need to be injected with data from inputs or outputs of modules.
    """

    attachment_name = "nnsight_module_nodes"
    condition: bool = False

    @classmethod
    def add(
        cls,
        graph: "Graph",
        proxy_value: Any,
        args: List[Any] = None,
        kwargs: Dict[str, Any] = None,
    ) -> Proxy:
        """Adds an InterventionProtocol Node to a Graph.

        Args:
            graph (Graph): Graph to add to.
            module_path (str): Module path of data this Node depends on (ex. model.module1.module2.output)
            proxy_value (Any): Proxy value.
            args (List[Any], optional): Args. Defaults to None.
            kwargs (Dict[str, Any], optional): Kwargs. Defaults to None.

        Returns:
            Proxy: _description_
        """

        # Creates the InterventionProtocol Node.
        proxy = graph.create(
            proxy_value=proxy_value, target=cls, args=args, kwargs=kwargs
        )

        cls.compile(proxy.node)

        return proxy

    @classmethod
    def compile(cls, node: Node) -> None:

        graph = node.graph

        module_path, *_ = node.args

        # Add attachment if it does not exist.
        if cls.attachment_name not in graph.attachments:

            graph.attachments[cls.attachment_name] = dict()

        # More than one Node can depend on a given input or output, therefore we store a list of node names.
        arguments = graph.attachments[cls.attachment_name]

        if module_path not in arguments:
            arguments[module_path] = []

        # Append the newly created nodes name.
        arguments[module_path].append(node.name)

    @classmethod
    def get_interventions(cls, graph: "Graph") -> Dict:
        """Returns mapping from module_paths to  InterventionNode names added to the given Graph.

        Args:
            graph (Graph): Graph.

        Returns:
            Dict: Interventions.
        """

        return graph.attachments.get(cls.attachment_name, dict())

    @classmethod
    def concat(
        cls,
        activations: Any,
        value: Any,
        batch_start: int,
        batch_size: int,
        total_batch_size: int,
    ):
        def _concat(values):

            data_type = type(values[0])

            if data_type == torch.Tensor:
                orig_size = values[-1]
                new_size = sum([value.shape[0] for value in values[:-1]])
                if new_size == orig_size:
                    return torch.concatenate(values[:-1])

                return values[0]
            elif data_type == list:
                return [
                    _concat([value[value_idx] for value in values])
                    for value_idx in range(len(values[0]))
                ]
            elif data_type == tuple:
                return tuple(
                    [
                        _concat([value[value_idx] for value in values])
                        for value_idx in range(len(values[0]))
                    ]
                )
            elif data_type == dict:
                return {
                    key: _concat([value[key] for value in values])
                    for key in values[0].keys()
                }
            return values[0]

        def narrow1(acts: torch.Tensor):
            if total_batch_size == acts.shape[0]:
                return acts.narrow(0, 0, batch_start)

            return acts

        pre = util.apply(activations, narrow1, torch.Tensor)

        post_batch_start = batch_start + batch_size

        def narrow2(acts: torch.Tensor):
            if total_batch_size == acts.shape[0]:
                return acts.narrow(
                    0, post_batch_start, acts.shape[0] - post_batch_start
                )

            return acts

        post = util.apply(
            activations,
            narrow2,
            torch.Tensor,
        )

        orig_sizes = util.apply(activations, lambda x: x.shape[0], torch.Tensor)

        return _concat([pre, value, post, orig_sizes])

    @classmethod
    def intervene(
        cls,
        activations: Any,
        module_path: str,
        key: str,
        intervention_handler: InterventionHandler,
    ):
        """Entry to intervention graph. This should be hooked to all modules involved in the intervention graph.

        Forms the current module_path key in the form of <module path>.<output/input>
        Checks the graphs InterventionProtocol attachment attribute for this key.
        If exists, value is a list of node names to iterate through.
        Node args for intervention type nodes should be ``[module_path, batch_size, batch_start, call_iter]``.
        Checks and updates the counter for the given intervention node. If counter is not ready yet continue.
        Using batch_size and batch_start, apply torch.narrow to tensors in activations to select
        only batch indexed tensors relevant to this intervention node. Sets the value of a node
        using the indexed values. Using torch.narrow returns a view of the tensors as opposed to a copy allowing
        subsequent downstream nodes to make edits to the values only in the relevant tensors, and have it update the original
        tensors. This both prevents interventions from effecting bathes outside their preview and allows edits
        to the output from downstream intervention nodes in the graph.

        Args:
            activations (Any): Either the inputs or outputs of a torch module.
            module_path (str): Module path of the current relevant module relative to the root model.
            key (str): Key denoting either "input" or "output" of module.
            intervention_handler (InterventionHandler): Handler object that stores the intervention graph and keeps track of module call count.

        Returns:
            Any: The activations, potentially modified by the intervention graph.
        """

        # Key to module activation intervention nodes has format: <module path>.<output/input>
        module_path = f"{module_path}.{key}"

        interventions = cls.get_interventions(intervention_handler.graph)

        if module_path in interventions:
            intervention_node_names = interventions[module_path]

            # Multiple intervention nodes can have same module_path if there are multiple invocations.
            for intervention_node_name in intervention_node_names:

                node = intervention_handler.graph.nodes[intervention_node_name]

                # Args for intervention nodes are (module_path, batch_group_idx, call_iter).
                _, batch_group_idx, call_iter = node.args

                batch_start, batch_size = intervention_handler.batch_groups[
                    batch_group_idx
                ]

                # Updates the count of intervention node calls.
                # If count matches call_iter, time to inject value into node.
                if call_iter != intervention_handler.count(
                    intervention_node_name
                ):

                    continue

                value = activations

                narrowed = False

                if len(intervention_handler.batch_groups) > 1:

                    def narrow(acts: torch.Tensor):

                        if acts.shape[0] == intervention_handler.batch_size:

                            nonlocal narrowed

                            narrowed = True

                            return acts.narrow(0, batch_start, batch_size)

                        return acts

                    value = util.apply(
                        activations,
                        narrow,
                        torch.Tensor,
                    )

                # Value injection.
                node.set_value(value)

                # Check if through the previous value injection, there was a 'swap' intervention.
                # This would mean we want to replace activations for this batch with some other ones.
                value = protocols.SwapProtocol.get_swap(
                    intervention_handler.graph, value
                )

                # If we narrowed any data, we need to concat it with data before and after it.
                if narrowed:

                    activations = cls.concat(
                        activations,
                        value,
                        batch_start,
                        batch_size,
                        intervention_handler.batch_size,
                    )
                # Otherwise just return the whole value as the activations.
                else:

                    activations = value

        return activations

    @classmethod
    def style(cls) -> Dict[str, Any]:
        """Visualization style for this protocol node.

        Returns:
            - Dict: dictionary style.
        """

        return {
            "node": {"color": "green4", "shape": "box"},  # Node display
            "label": cls.__name__,
            "arg": defaultdict(
                lambda: {"color": "gray", "shape": "box"}
            ),  # Non-node argument display
            "arg_kname": defaultdict(
                lambda: None, {0: "key", 1: "batch_size", 2: "batch_start"}
            ),  # Argument label key word
            "edge": defaultdict(lambda: "solid"),
        }  # Argument Edge display


class HookHandler(AbstractContextManager):
    """Context manager that applies input and/or output hooks to modules in a model.

    Registers provided hooks on __enter__ and removes them on __exit__.

    Attributes:
        model (torch.nn.Module): Root model to access modules and apply hooks to.
        modules (List[Tuple[torch.nn.Module, str]]): List of modules to apply hooks to along with their module_path.
        input_hook (Callable): Function to apply to inputs of designated modules.
            Should have signature of [inputs(Any), module_path(str)] -> inputs(Any)
        output_hook (Callable): Function to apply to outputs of designated modules.
            Should have signature of [outputs(Any), module_path(str)] -> outputs(Any)
        handles (List[RemovableHandle]): Handles returned from registering hooks as to be used when removing hooks on __exit__.
    """

    def __init__(
        self,
        model: torch.nn.Module,
        module_keys: List[str],
        input_hook: Callable = None,
        output_hook: Callable = None,
    ) -> None:
        self.model = model
        self.module_keys = module_keys

        self.input_hook = input_hook
        self.output_hook = output_hook

        self.handles: List[RemovableHandle] = []

    def __enter__(self) -> HookHandler:
        """Registers input and output hooks to modules if they are defined.

        Returns:
            HookModel: HookModel object.
        """

        for module_key in self.module_keys:

            module_atoms = module_key.split(".")

            if len(module_atoms) == 1:
                continue

            *module_atoms, hook_type = module_atoms

            module_path = ".".join(module_atoms)

            module: torch.nn.Module = util.fetch_attr(self.model, module_path)

            if hook_type == "input":

                def input_hook(module, input, kwargs, module_path=module_path):
                    return self.input_hook((input, kwargs), module_path)

                self.handles.append(
                    module.register_forward_pre_hook(
                        input_hook, with_kwargs=True, prepend=True
                    )
                )

            elif hook_type == "output":

                def output_hook(module, input, output, module_path=module_path):
                    return self.output_hook(output, module_path)

                self.handles.append(
                    module.register_forward_hook(output_hook, prepend=True)
                )

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Removes all handles added during __enter__."""

        for handle in self.handles:
            handle.remove()

        if isinstance(exc_val, Exception):
            raise exc_val


class InterventionHandler:
    """Object passed to InterventionProtocol.intervene to store information about the current interleaving execution run.

    Like the Intervention Graph, the total batch size that is being executed, and a counter for how many times an Intervention node has been attempted to be executed.
    """

    def __init__(
        self, graph: Graph, batch_groups: List[Tuple[int, int]], batch_size: int
    ) -> None:

        self.graph = graph
        self.batch_groups = batch_groups
        self.batch_size = batch_size
        self.call_counter: Dict[str, int] = {}

    def count(self, name: str) -> int:
        """Increments the count of times a given Intervention Node has tried to be executed and returns the count.

        Args:
            name (str): Name of intervention node to return count for.

        Returns:
            int: Count.
        """

        if name not in self.call_counter:

            self.call_counter[name] = 0

        else:

            self.call_counter[name] += 1

        return self.call_counter[name]



from __future__ import annotations

import json
import warnings
from typing import Any, Dict, List, Optional, Tuple, Type, Union

import torch
from transformers import (
    AutoConfig,
    AutoModel,
    AutoModelForCausalLM,
    AutoTokenizer,
    BatchEncoding,
    PreTrainedModel,
    PreTrainedTokenizer,
)
from transformers.models.auto import modeling_auto
from transformers.models.llama.configuration_llama import LlamaConfig
from typing_extensions import Self

from nnsight.envoy import Envoy

from ..intervention import InterventionProxy
from ..util import WrapperModule
from . import NNsight
from .mixins import GenerationMixin, RemoteableMixin


class TokenIndexer:
    """Helper class to directly access token indices of hidden states.
    Directly indexes the second dimension of tensors.
    Makes positive indices negative as tokens are padded on the left.

    Args:
        proxy (InterventionProxy): Proxy to aid in token indexing.
    """

    def __init__(self, proxy: InterventionProxy) -> None:
        self.proxy = proxy

    def convert_idx(self, idx: int):
        if idx >= 0:
            n_tokens = self.proxy.node.proxy_value.shape[1]
            idx = -(n_tokens - idx)

        return idx

    def __getitem__(self, key: int) -> LanguageModelProxy:
        key = self.convert_idx(key)

        return self.proxy[:, key]

    def __setitem__(self, key: int, value: Union[LanguageModelProxy, Any]) -> None:
        key = self.convert_idx(key)

        self.proxy[:, key] = value


class LanguageModelProxy(InterventionProxy):
    """

    Indexing by token of hidden states can easily done using ``.token[<idx>]`` or ``.t[<idx>]``

    .. code-block:: python

        with runner.invoke('The Eiffel Tower is in the city of') as invoker:
            logits = model.lm_head.output.t[0].save()

        print(logits.value)

    This would save only the first token of the output for this module.
    This should be used when using multiple invokes as the batching and padding of multiple inputs could mean the indices for tokens shifts around and this take care of that.

    Args:
        InterventionProxy (_type_): _description_

    Returns:
        _type_: _description_
    """

    @property
    def token(self) -> TokenIndexer:
        """Property used to do token based indexing on a proxy.
        Directly indexes the second dimension of tensors.
        Makes positive indices negative as tokens are padded on the left.

        Example:

            .. code-block:: python

                model.transformer.h[0].mlp.output.token[0]

            Is equivalent to:

            .. code-block:: python

                model.transformer.h[0].mlp.output.token[:,-3]

            For a proxy tensor with 3 tokens.

        Returns:
            TokenIndexer: Object to do token based indexing.
        """
        return TokenIndexer(self)

    @property
    def t(self) -> TokenIndexer:
        """Property as alias for InterventionProxy.token"""
        return self.token


class LanguageModel(GenerationMixin, RemoteableMixin, NNsight):
    """LanguageModels are NNsight wrappers around transformers language models.

    Inputs can be in the form of:
        Prompt: (str)
        Prompts: (List[str])
        Batched prompts: (List[List[str]])
        Tokenized prompt: (Union[List[int], torch.Tensor])
        Tokenized prompts: (Union[List[List[int]], torch.Tensor])
        Direct input: (Dict[str,Any])

    If using a custom model, you also need to provide the tokenizer like ``LanguageModel(custom_model, tokenizer=tokenizer)``

    Calls to generate pass arguments downstream to :func:`GenerationMixin.generate`

    Attributes:
        config (PretrainedConfig): Huggingface config file loaded from repository or checkpoint.
        tokenizer (PreTrainedTokenizer): Tokenizer for LMs.
        automodel (Type): AutoModel type from transformer auto models.
        model (PreTrainedModel): Meta version of underlying auto model.

    """

    proxy_class = LanguageModelProxy

    def __new__(cls, *args, **kwargs) -> Self | Envoy:
        return object.__new__(cls)

    def __init__(
        self,
        model_key: Union[str, torch.nn.Module],
        *args,
        tokenizer: Optional[PreTrainedTokenizer] = None,
        automodel: Type[AutoModel] = AutoModelForCausalLM,
        **kwargs,
    ) -> None:
        self.tokenizer: PreTrainedTokenizer = tokenizer
        self._model: PreTrainedModel = None
        self.automodel = (
            automodel
            if not isinstance(automodel, str)
            else getattr(modeling_auto, automodel)
        )

        if isinstance(model_key, torch.nn.Module):

            setattr(model_key, "generator", WrapperModule())

        super().__init__(model_key, *args, **kwargs)

    def _load(
        self,
        repo_id: str,
        tokenizer_kwargs: Optional[Dict[str, Any]] = None,
        patch_llama_scan: bool = True,
        **kwargs,
    ) -> PreTrainedModel:

        config = kwargs.pop("config", None) or AutoConfig.from_pretrained(
            repo_id, **kwargs
        )

        if self.tokenizer is None:
            if tokenizer_kwargs is None:
                tokenizer_kwargs = {}

            self.tokenizer = AutoTokenizer.from_pretrained(
                repo_id, config=config, **tokenizer_kwargs
            )

            if not hasattr(self.tokenizer.pad_token, "pad_token"):
                self.tokenizer.pad_token = self.tokenizer.eos_token

            if (
                patch_llama_scan
                and isinstance(config, LlamaConfig)
                and isinstance(config.rope_scaling, dict)
                and "rope_type" in config.rope_scaling
            ):
                config.rope_scaling["rope_type"] = "default"

            model = self.automodel.from_config(config, trust_remote_code=True)

            setattr(model, "generator", WrapperModule())

            return model

        if (
            patch_llama_scan
            and isinstance(config, LlamaConfig)
            and isinstance(config.rope_scaling, dict)
            and "rope_type" in config.rope_scaling
        ):
            config.rope_scaling["rope_type"] = "llama3"

        model = self.automodel.from_pretrained(repo_id, config=config, **kwargs)

        setattr(model, "generator", WrapperModule())

        return model

    def _tokenize(
        self,
        inputs: Union[
            str,
            List[str],
            List[List[str]],
            List[int],
            List[List[int]],
            torch.Tensor,
            Dict[str, Any],
        ],
        **kwargs,
    ):
        if isinstance(inputs, BatchEncoding):
            return inputs

        if isinstance(inputs, str) or (
            isinstance(inputs, list) and isinstance(inputs[0], int)
        ):
            inputs = [inputs]

        if isinstance(inputs, torch.Tensor) and inputs.ndim == 1:
            inputs = inputs.unsqueeze(0)

        if not isinstance(inputs[0], str):
            inputs = [{"input_ids": ids} for ids in inputs]
            return self.tokenizer.pad(inputs, return_tensors="pt", **kwargs)

        return self.tokenizer(inputs, return_tensors="pt", padding=True, **kwargs)

    def _prepare_inputs(
        self,
        inputs: Union[
            str,
            List[str],
            List[List[str]],
            List[int],
            List[List[int]],
            torch.Tensor,
            Dict[str, Any],
            BatchEncoding,
        ],
        labels: Any = None,
        **kwargs,
    ) -> Tuple[BatchEncoding, int]:
        if isinstance(inputs, dict):

            new_inputs = dict()

            tokenized_inputs = self._tokenize(inputs["input_ids"], **kwargs)

            new_inputs["input_ids"] = tokenized_inputs["input_ids"]

            if "attention_mask" in inputs:
                for ai, attn_mask in enumerate(inputs["attention_mask"]):
                    tokenized_inputs["attention_mask"][
                        ai, -len(attn_mask) :
                    ] = attn_mask

                new_inputs["attention_mask"] = tokenized_inputs["attention_mask"]

            if "labels" in inputs:
                labels = self._tokenize(inputs["labels"], **kwargs)

                new_inputs["labels"] = labels["input_ids"]

            return (BatchEncoding(new_inputs),), len(new_inputs["input_ids"])

        inputs = self._tokenize(inputs, **kwargs)

        if labels is not None:
            labels = self._tokenize(labels, **kwargs)

            inputs["labels"] = labels["input_ids"]

        return (inputs,), len(inputs["input_ids"])

    def _batch_inputs(
        self,
        batched_inputs: Optional[Dict[str, Any]],
        prepared_inputs: BatchEncoding,
    ) -> Tuple[Dict[str, Any]]:

        if batched_inputs is None:
            batched_inputs = {"input_ids": []}

            if "labels" in prepared_inputs:
                batched_inputs["labels"] = []

            if "attention_mask" in prepared_inputs:
                batched_inputs["attention_mask"] = []

        else:

            batched_inputs = batched_inputs[0]

        batched_inputs["input_ids"].extend(prepared_inputs["input_ids"])

        if "labels" in prepared_inputs:
            batched_inputs["labels"].extend(prepared_inputs["labels"])
        if "attention_mask" in prepared_inputs:
            batched_inputs["attention_mask"].extend(prepared_inputs["attention_mask"])

        return (batched_inputs,)

    def _execute_forward(self, prepared_inputs: Any, *args, **kwargs):

        device = next(self._model.parameters()).device

        return self._model(
            *args,
            **prepared_inputs.to(device),
            **kwargs,
        )

    def _execute_generate(
        self, prepared_inputs: Any, *args, max_new_tokens=1, **kwargs
    ):

        device = next(self._model.parameters()).device

        output = self._model.generate(
            *args,
            **prepared_inputs.to(device),
            max_new_tokens=max_new_tokens,
            **kwargs,
        )

        self._model.generator(output)

        return output

    def _remoteable_model_key(self) -> str:
        return json.dumps(
            {"repo_id": self._model_key}  # , "torch_dtype": str(self._model.dtype)}
        )

    @classmethod
    def _remoteable_from_model_key(cls, model_key: str, **kwargs) -> Self:

        kwargs = {**json.loads(model_key), **kwargs}

        repo_id = kwargs.pop("repo_id")

        return LanguageModel(repo_id, **kwargs)



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Dict, List, Union

from pydantic import BaseModel, ConfigDict

from .. import NNsight
from .format.types import *

if TYPE_CHECKING:
    from ..contexts.backends.RemoteBackend import RemoteMixin


class RequestModel(BaseModel):
    
    model_config = ConfigDict(arbitrary_types_allowed=True, protected_namespaces=())

    object: Union[SessionType, TracerType, SessionModel, TracerModel]
    model_key: str

    id: str = None
    received: datetime = None

    session_id: Optional[str] = None

    def deserialize(self, model: NNsight) -> "RemoteMixin":
        
        handler = DeserializeHandler(model=model)

        return self.object.deserialize(handler)



from __future__ import annotations

import inspect
import tempfile
from typing import Callable, Dict, Optional, Type

from PIL import Image as PILImage

from .. import util
from ..util import apply
from .Node import Node
from .protocols import EarlyStopProtocol, Protocol
from .Proxy import Proxy
from .util import validate


class Graph:
    """Represents a computation graph composed of :class:`Nodes <nnsight.tracing.Node.Node>`.


    Attributes:
        nodes (Dict[str, :class:`Node <nnsight.tracing.Node.Node>`]): Mapping of `Node` name to node. Order is preserved and important when executing the graph sequentially.
        attachments (Dict[str, Any]): Dictionary object used to add extra functionality to this Graph. Used by Protocols.
        proxy_class (Type[class:`Proxy <nnsight.tracing.Proxy.Proxy>`]): Proxy class to use. Defaults to class:`Proxy <nnsight.tracing.Proxy.Proxy>`.
        alive (bool): If this Graph should be considered alive (still tracing), and therefore added to. Used by `Node`s.
        name_idx (Dict[str, int]): Mapping of node target_name to number of previous names with the same target_name. Used so names are unique.
        
        validate (bool): If to execute nodes as they are added with their proxy values in order to check if the executions are possible and create a new proxy_value. Defaults to True.
        
            When adding `Node`s to the `Graph`, if the `Graph`'s validate attribute is set to `True`, \
            it will execute the `Node`'s target with its arguments' `.proxy_value` attributes (essentially executing the Node, with FakeTensors in FakeTensorMode).
            This 1.) checks to see of the operation is valid on the tensor shape's within the `.proxy_value`s (this would catch an indexing error) and \
            2.) populating this new `Node`'s `.proxy_value` attribute with the result.
        
        sequential (bool): If to run nodes sequentially when executing this graph.
        
            When this is set to `True`, `Node`s attempt to be executed in the order they were added to the `Graph` when calling `.execute(). \
            Otherwise, all nodes are checked to be fulfilled (they have no dependencies). These are root nodes and they are then executed in the order they were added.

        
    """

    def __init__(
        self,
        proxy_class: Type[Proxy] = Proxy,
        validate: bool = False,
        sequential: bool = True,
        graph_id: int = None,
    ) -> None:

        self.id = graph_id or id(self)

        self.proxy_class = proxy_class
        self.validate = validate
        self.sequential = sequential

        self.alive = True

        self.nodes: Dict[str, Node] = dict()
        self.name_idx: Dict[str, int] = dict()

        self.attachments = dict()

    def reset(self) -> None:
        """Resets the Graph to prepare for a new execution of the Graph.
        Calls `.reset()` on all Nodes.
        """

        # Reset Nodes individually.
        for node in self.nodes.values():
            node.reset()

    def execute(self) -> None:
        """Executes operations of `Graph`.

        Executes all `Node`s sequentially if `Graph.sequential`. Otherwise execute only root `Node`s sequentially.
        """

        if self.sequential:
            is_stopped_early: bool = False
            early_stop_execption: Optional[
                EarlyStopProtocol.EarlyStopException
            ] = None
            for node in self.nodes.values():
                if not is_stopped_early:
                    if node.fulfilled():
                        try:
                            node.execute()
                        except EarlyStopProtocol.EarlyStopException as e:
                            is_stopped_early = True
                            early_stop_execption = e
                            continue
                else:
                    node.clean()
            if is_stopped_early:
                raise early_stop_execption
        else:

            root_nodes = [
                node for node in self.nodes.values() if node.fulfilled()
            ]

            for node in root_nodes:
                node.execute()

    def create(self, *args, **kwargs) -> Proxy:
        """Creates a Node directly on this `Graph` and returns its `Proxy`.

        Returns:
            Proxy: `Proxy` for newly created `Node`.
        """

        return self.proxy_class(Node(*args, graph=self, **kwargs))

    def add(self, node: Node) -> None:
        """Adds a Node to this Graph. Called by Nodes on __init__.
        
        When adding `Node`s to the `Graph`, if the `Graph`'s validate attribute is set to `True`, \
        it will execute the `Node`'s target with its arguments' `.proxy_value` attributes (essentially executing the Node, with FakeTensors in FakeTensorMode).
        This 1.) checks to see of the operation is valid on the tensor shape's within the `.proxy_value`s (this would catch an indexing error) and \
        2.) populating this new `Node`'s `.proxy_value` attribute with the result.
    

        Args:
            node (Node): Node to add.
        """

        # If we're validating and the user did not provide a proxy_value, execute the given target with meta proxy values to compute new proxy_value.
        if self.validate and node.proxy_value is inspect._empty:

            node.proxy_value = validate(node.target, *node.args, **node.kwargs)

        # Get name of target.
        name = (
            node.target
            if isinstance(node.target, str)
            else node.target.__name__
        )

        # Init name_idx tracker for this Node's name if not already added.
        if name not in self.name_idx:
            self.name_idx[name] = 0

        # If Node's name is not set, set it to the name_idxed version.
        if node.name is None:
            node.name = f"{name}_{self.name_idx[name]}"

        # Increment name_idx for name.
        self.name_idx[name] += 1

        # Add Node.
        self.nodes[node.name] = node

    def copy(self):
        """Copy constructs a new Graph and then recursively
        creates new Nodes on the graph.
        """
        new_graph = Graph(
            validate=self.validate,
            sequential=self.sequential,
            proxy_class=self.proxy_class,
        )

        def compile(graph: Graph, old_node: Node):
            if old_node.name in graph.nodes:
                return graph.nodes[old_node.name]

            node = graph.create(
                target=old_node.target,
                name=old_node.name,
                proxy_value=None,
                args=apply(old_node.args, lambda x: compile(graph, x), Node),
                kwargs=apply(
                    old_node.kwargs, lambda x: compile(graph, x), Node
                ),
            ).node

            if isinstance(node.target, type) and issubclass(
                node.target, Protocol
            ):
                node.target.compile(node)

            return node

        # To preserve order
        nodes = {}

        for node in self.nodes.values():

            compile(new_graph, node)

            # To preserve order
            nodes[node.name] = new_graph.nodes[node.name]

        # To preserve order
        new_graph.nodes = nodes

        return new_graph

    def vis(
        self,
        title: str = "graph",
        path: str = ".",
        display: bool = True,
        save: bool = False,
        recursive: bool = False,
    ):
        """Generates and saves a graphical visualization of the Intervention Graph using the pygraphviz library.
        Args:
            title (str): Name of the Intervention Graph. Defaults to "graph".
            path (str): Directory path to save the graphic in. If None saves content to the current directory.
            display (bool): If True, shows the graph image.
            save (bool): If True, saves the graph to the specified path.
            recursive (bool): If True, recursively visualize sub-graphs.
        """

        try:

            import pygraphviz as pgv

        except Exception as e:

            raise type(e)(
                "Visualization of the Graph requires `pygraphviz` which requires `graphviz` to be installed on your machine."
            ) from e

        from IPython.display import Image
        from IPython.display import display as IDisplay

        graph: pgv.AGraph = pgv.AGraph(strict=True, directed=True)

        graph.graph_attr.update(
            label=title, fontsize="20", labelloc="t", labeljust="c"
        )

        for node in self.nodes.values():
            # draw bottom up
            if len(node.listeners) == 0:
                node.visualize(graph, recursive)

        def display_graph(file_name):
            in_notebook = True

            # Credit: Till Hoffmann - https://stackoverflow.com/a/22424821
            try:
                from IPython import get_ipython

                if "IPKernelApp" not in get_ipython().config:
                    in_notebook = False
            except ImportError:
                in_notebook = False
            except AttributeError:
                in_notebook = False

            if in_notebook:
                IDisplay(Image(filename=file_name))
            else:
                img = PILImage.open(file_name)
                img.show()
                img.close()

        if not save:
            with tempfile.NamedTemporaryFile(suffix=".png") as temp_file:
                graph.draw(temp_file.name, prog="dot")
                if display:
                    display_graph(temp_file.name)
        else:
            graph.draw(f"{path}/{title}.png", prog="dot")
            if display:
                display_graph(f"{path}/{title}.png")

    def __str__(self) -> str:
        result = ""

        for name, node in self.nodes.items():
            result += f"  %{node}\n"

        return result



import torch
from abc import ABC, abstractmethod

from nnsight import LanguageModel, Module
from typing import Any, Callable, Dict, List, Union

from functools import reduce

import torch

def shift_preds(x: torch.Tensor, shift: int):
    """Shift predictions by a given amount.

    Args:
        x: (batch x seq_len) predictions to shift.
        shift: Amount to shift by. Positive values take from the end, negative values
            from the start.

    Returns:
        (batch x (seq_len - shift)) predictions shifted by the given amount.
    """
    if shift > 0:
        return x[:, :-shift]
    if shift < 0:
        return x[:, -shift:]

    return x

# implement kl loss

    
    
    
    


from __future__ import annotations

from typing import Any, Callable, Dict, List, Optional, Union

import torch
from diffusers import DiffusionPipeline
from transformers import BatchEncoding
from typing_extensions import Self

from .. import util
from ..envoy import Envoy
from .mixins import GenerationMixin
from .NNsightModel import NNsight


class Diffuser(util.WrapperModule):
    def __init__(self, *args, **kwargs) -> None:
        super().__init__()

        self.pipeline = DiffusionPipeline.from_pretrained(*args, **kwargs)

        for key, value in self.pipeline.__dict__.items():
            if isinstance(value, torch.nn.Module):
                setattr(self, key, value)

        self.tokenizer = self.pipeline.tokenizer


class DiffusionModel(GenerationMixin, NNsight):

    def __new__(cls, *args, **kwargs) -> Self | Envoy:
        return object.__new__(cls)

    def __init__(self, *args, **kwargs) -> None:

        self._model: Diffuser = None

        super().__init__(*args, **kwargs)

    def _load(self, repo_id: str, device_map=None, **kwargs) -> Diffuser:

        if self._model is None:

            model = Diffuser(
                repo_id,
                device_map=None,
                low_cpu_mem_usage=False,
                **kwargs,
            )

            return model

        model = Diffuser(repo_id, device_map=device_map, **kwargs)

        return model

    def _prepare_inputs(
        self,
        inputs: Union[str, List[str]],
    ) -> Any:

        if isinstance(inputs, str):
            inputs = [inputs]

        return (inputs,), len(inputs)

    def _batch_inputs(
        self,
        batched_inputs: Optional[Dict[str, Any]],
        prepared_inputs: BatchEncoding,
    ) -> torch.Tensor:

        if batched_inputs is None:

            return prepared_inputs

        return batched_inputs + prepared_inputs

    def _execute_forward(self, prepared_inputs: Any, *args, **kwargs):

        return self._model.unet(
            prepared_inputs,
            *args,
            **kwargs,
        )

    def _execute_generate(
        self, prepared_inputs: Any, *args, seed: int = None, **kwargs
    ):

        if self._scanning():

            kwargs["num_inference_steps"] = 1

        generator = torch.Generator()

        if seed is not None:

            generator = generator.manual_seed(seed)

        output = self._model.pipeline(
            prepared_inputs, *args, generator=generator, **kwargs
        )

        output = self._model(output)

        return output



from __future__ import annotations

import weakref
from typing import TYPE_CHECKING, Any, Dict, Iterable

from ...tracing.Bridge import Bridge
from ...tracing.Graph import Graph
from ...tracing.protocols import EarlyStopProtocol
from ..backends import Backend, BridgeBackend, RemoteMixin
from ..GraphBasedContext import GraphBasedContext
from .Iterator import Iterator

if TYPE_CHECKING:
    from ...models.mixins import RemoteableMixin
    from ...models.NNsightModel import NNsight


class Session(GraphBasedContext, RemoteMixin):
    """A Session is a root Collection that handles adding new Graphs and new Collections while in the session.

    Attributes:
        bridge (Bridge): Bridge object which stores all Graphs added during the session and handles interaction between them
        graph (Graph): Root Graph where operations and values meant for access by all subsequent Graphs should be stored and referenced.
        model (NNsight): NNsight model.
        backend (Backend): Backend for this context object.
    """

    def __init__(
        self,
        backend: Backend,
        model: "NNsight",
        *args,
        bridge: Bridge = None,
        **kwargs,
    ) -> None:

        self.bridge = Bridge() if bridge is None else bridge

        self.model = model

        GraphBasedContext.__init__(
            self,
            backend,
            bridge=self.bridge,
            proxy_class=self.model.proxy_class,
            *args,
            **kwargs,
        )

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:

        self.model._session = None

        super().__exit__(exc_type, exc_val, exc_tb)

    def iter(self, iterable: Iterable, **kwargs) -> Iterator:
        """Creates an Iterator context to iteratively execute an intervention graph, with an update item at each iteration.

        Args:
            - iterable (Iterable): Data to iterate over.
            - return_context (bool): If True, returns the Iterator context. Default: False.

        Returns:
            Iterator: Iterator context.

        Example:
            Setup:
                .. code-block:: python
                    import torch
                    from collections import OrderedDict
                    input_size = 5
                    hidden_dims = 10
                    output_size = 2
                    model = nn.Sequential(OrderedDict([
                        ('layer1', torch.nn.Linear(input_size, hidden_dims)),
                        ('layer2', torch.nn.Linear(hidden_dims, output_size)),
                    ]))
                    input = torch.rand((1, input_size))

            Ex:
                .. code-block:: python
                    with model.session() as session:
                        l  = session.apply(list).save()
                        with session.iter([0, 1, 2]) as item:
                            l.append(item)
        """

        bridge = weakref.proxy(self.bridge)

        backend = BridgeBackend(bridge)

        return Iterator(
            iterable,
            backend,
            bridge=bridge,
            proxy_class=self.model.proxy_class,
            **kwargs,
        )

    ### BACKENDS ########

    def local_backend_execute(self) -> Dict[int, Graph]:

        try:
            super().local_backend_execute()
        except EarlyStopProtocol.EarlyStopException:
            pass

        local_result = self.bridge.id_to_graph

        self.bridge = weakref.proxy(self.bridge)

        return local_result

    def remote_backend_get_model_key(self) -> str:

        self.model: "RemoteableMixin"

        return self.model.to_model_key()

    def remote_backend_postprocess_result(self, local_result: Dict[int, Graph]):

        from ...schema.Response import ResultModel

        return {
            id: ResultModel.from_graph(graph)
            for id, graph in local_result.items()
        }

    def remote_backend_handle_result_value(
        self, value: Dict[int, Dict[str, Any]]
    ):

        for graph_id, saves in value.items():

            graph = self.bridge.id_to_graph[graph_id]

            for node_name, node_value in saves.items():
                graph.nodes[node_name]._value = node_value

            graph.alive = False

    def remote_backend_cleanup(self):

        self.bridge = weakref.proxy(self.bridge)

        graph = self.graph
        graph.alive = False

        if not isinstance(graph, weakref.ProxyType):
            self.graph = weakref.proxy(graph)

    def __repr__(self) -> str:
        return f"&lt;{self.__class__.__name__} at {hex(id(self))}&gt;"



import torch

def sigmoid_boundary(_input, boundary_x, boundary_y, temperature):
    """Generate sigmoid mask"""
    return torch.sigmoid((_input - boundary_x) / temperature) * \
        torch.sigmoid((boundary_y - _input) / temperature)


from typing import Any

from ... import NNsight
from ...contexts.Tracer import Tracer

class GenerationMixin(NNsight):

    def generate(self, *args, **kwargs) -> Tracer:

        return self.trace(*args, generate=True, **kwargs)

    def _execute(
        self, *args, generate: bool = False, **kwargs
    ) -> Any:

        if generate:

            return self._execute_generate(*args, **kwargs)

        return self._execute_forward(*args, **kwargs)

    def _scan(
        self, prepared_inputs: Any, *args, generate: bool = False, **kwargs
    ) -> Any:

        if generate:

            return self._scan_generate(prepared_inputs, *args, **kwargs)

        return self._scan_forward(prepared_inputs, *args, **kwargs)

    def _execute_forward(self, prepared_inputs: Any, *args, **kwargs):

        raise NotImplementedError()

    def _execute_generate(self, prepared_inputs: Any, *args, **kwargs):

        raise NotImplementedError()

    def _scan_forward(self, prepared_inputs: Any, *args, **kwargs):

        raise NotImplementedError()

    def _scan_generate(self, prepared_inputs: Any, *args, **kwargs):

        raise NotImplementedError()



from __future__ import annotations

import copy
from contextlib import AbstractContextManager
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Tuple

import torch
from torch._subclasses.fake_tensor import FakeCopyMode, FakeTensorMode
from torch.fx.experimental.symbolic_shapes import ShapeEnv

from .. import util
from ..patching import Patch, Patcher
from ..tracing.Node import Node
from ..tracing.Proxy import Proxy
from . import check_for_dependencies
from .GraphBasedContext import GlobalTracingContext

if TYPE_CHECKING:

    from .Tracer import Tracer


class Invoker(AbstractContextManager):
    """An Invoker is meant to work in tandem with a :class:`nnsight.contexts.Tracer.Tracer` to enter input and manage intervention tracing.

    Attributes:
        tracer (nnsight.contexts.Tracer.Tracer): Tracer object to enter input and manage context.
        inputs (tuple[Any]): Initially entered inputs, then post-processed inputs from model's ._prepare_inputs(...) method.
        scan (bool): If to execute the model using `FakeTensor` in order to update the potential sizes/dtypes of all modules' Envoys' inputs/outputs as well as validate things work correctly.
            Scanning is not free computation wise so you may want to turn this to false when running in a loop.
            When making interventions, you made get shape errors if scan is false as it validates operations based on shapes so
            for looped calls where shapes are consistent, you may want to have scan=True for the first loop. Defaults to True.
        kwargs (Dict[str,Any]): Keyword arguments passed to the model's _prepare_inputs method.
        scanning (bool): If currently scanning.
    """

    def __init__(
        self,
        tracer: "Tracer",
        *inputs: Any,
        scan: bool = False,
        **kwargs,
    ) -> None:

        self.tracer = tracer
        self.inputs = inputs
        self.scan = scan
        self.kwargs = kwargs

        self.scanning = False

        self.tracer.invoker = self

    def __enter__(self) -> Invoker:
        """Enters a new invocation context with a given input.

        Calls the model's _prepare_inputs method using the input and other arguments.
        If scan is True, uses the model's ._execute method to update and validate module Envoy's inputs/outputs using a fake mode.
        Gets a batched version of the post processed input using the model's ._batched_inputs method to update the Tracer's
            current batch_size and batched_input.

        Returns:
            Invoker: Invoker.
        """

        has_proxies_in_inputs = False

        # If were accumulating, we might have Proxies in the input.
        # Therefore we first: Check to see if there are any Proxies.
        # If there are, preserve the raw inputs with Proxies converted to a Locked Bridge protocol.
        # Set self.inputs to be the proxy_value so we can prepare_inputs, get the batch size, and scan.
        if self.tracer.model._session is not None:

            self.inputs, has_proxies_in_inputs = check_for_dependencies(
                self.inputs
            )

        with GlobalTracingContext.exit_global_tracing_context():

            if not has_proxies_in_inputs:

                self.inputs, batch_size = self.tracer.model._prepare_inputs(
                    *self.inputs, **self.kwargs
                )

            if self.scan:

                inputs = self.inputs

                if has_proxies_in_inputs:

                    inputs = util.apply(inputs, lambda x: x.proxy_value, Node)

                    inputs, batch_size = self.tracer.model._prepare_inputs(
                        *inputs, **self.kwargs
                    )

                self.tracer.model._envoy._clear()

                self.scanning = True

                with Patcher() as patcher:

                    # Some logic (like gpt-j rotary embeddings) gets "poisoned" by FakeTensors.
                    # This does not happen when `torch._jit_internal.is_scripting() returns True.`
                    patcher.add(
                        Patch(torch._jit_internal, lambda: True, "is_scripting")
                    )

                    with FakeTensorMode(
                        allow_non_fake_inputs=True,
                        shape_env=ShapeEnv(assume_static_by_default=True),
                    ) as fake_mode:
                        with FakeCopyMode(fake_mode):
                            self.tracer.model._execute(
                                *copy.deepcopy(inputs),
                                **copy.deepcopy(self.tracer._kwargs),
                            )

                self.scanning = False

            else:
                self.tracer.model._envoy._reset()

            self.tracer._invoker_inputs.append(self.inputs)

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:

        self.tracer.invoker = None

        if isinstance(exc_val, BaseException):
            raise exc_val



from __future__ import annotations

import inspect
import warnings
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Union

import torch

from .contexts.backends import EditBackend
from .contexts.Tracer import Tracer
from .intervention import InterventionProtocol, InterventionProxy
from .tracing import protocols


class Envoy:
    """Envoy object act as proxies for torch modules within a model's module tree in order to add nnsight functionality.
    Proxies of the underlying module's output and input are accessed by `.output` and `.input` respectively.

    Attributes:
        path (str): String representing the attribute path of this Envoy's module relative the the root model. Separated by '.' e.x ('transformer.h.0.mlp'). Set by NNsight on initialization of meta model.
        _fake_outputs (List[torch.Tensor]): List of 'meta' tensors built from the outputs most recent _scan. Is list as there can be multiple shapes for a module called more than once.
        _fake_inputs (List[torch.Tensor]): List of 'meta' tensors built from the inputs most recent _scan. Is list as there can be multiple shapes for a module called more than once.
        output (nnsight.intervention.InterventionProxy): Proxy object representing the output of this Envoy's module. Reset on forward pass.
        input (nnsight.intervention.InterventionProxy): Proxy object representing the input of this Envoy's module. Reset on forward pass.
        _call_iter (int): Integer representing the current iteration of this Envoy's module's inputs/outputs.
        _tracer (nnsight.context.Tracer.Tracer): Object which adds this Envoy's module's output and input proxies to an intervention graph. Must be set on Envoys objects manually by the Tracer.
    """

    def __init__(self, module: torch.nn.Module, module_path: str = ""):

        self.path = module_path

        self._fake_outputs: List[torch.Tensor] = []
        self._fake_inputs: List[torch.Tensor] = []

        self._output: Optional[InterventionProxy] = None
        self._input: Optional[InterventionProxy] = None

        self._call_iter = 0

        self._tracer: Tracer = None

        self._module = module
        self._sub_envoys: List[Envoy] = []

        # Register hook on underlying module to update the _fake_outputs and _fake_inputs on forward pass.
        self._hook_handle = self._module.register_forward_hook(
            self._hook, with_kwargs=True
        )

        for name, module in self._module.named_children():

            setattr(self, name, module)

    def _update(self, module: torch.nn.Module) -> None:
        """Updates the ._model attribute using a new model of the same architecture.
        Used when loading the real weights (dispatching) and need to replace the underlying modules.
        """

        self._module = module

        self._hook_handle.remove()

        self._hook_handle = self._module.register_forward_hook(
            self._hook, with_kwargs=True
        )

        for i, module in enumerate(self._module.children()):

            self._sub_envoys[i]._update(module)

    def _add_envoy(self, module: torch.nn.Module, name: str) -> None:
        """Adds a new Envoy for a given torch module under this Envoy.

        Args:
            module (torch.nn.Module): Module to create Envoy for.
            name (str): name of envoy/attribute.
        """

        envoy = Envoy(module, module_path=f"{self.path}.{name}")

        self._sub_envoys.append(envoy)

        # If the module already has a sub-module named 'input' or 'output',
        # mount the proxy access to 'nns_input' or 'nns_output instead.
        if hasattr(Envoy, name):

            self._handle_overloaded_mount(envoy, name)

        else:

            super().__setattr__(name, envoy)

    def _handle_overloaded_mount(self, envoy: Envoy, mount_point: str) -> None:
        """If a given module already has an attribute of the same name as something nnsight wants to add, we need to rename it.

        Directly edits the underlying class to accomplish this.

        Args:
            envoy (Envoy): Envoy to handle.
            mount_point (str): Overloaded attribute name.
        """

        warnings.warn(
            f"Module of type `{type(self._module)}` has pre-defined a `{mount_point}` attribute. nnsight access for `{mount_point}` will be mounted at `.nns_{mount_point}` instead of `.{mount_point}` for this module only."
        )

        # If we already shifted a mount point dont create another new class.
        if "Preserved" in self.__class__.__name__:

            new_cls = self.__class__

        else:

            new_cls = type(
                f"{Envoy.__name__}.Preserved",
                (Envoy,),
                {},
            )

        # Get the normal proxy mount point
        mount = getattr(new_cls, mount_point)

        # Move it to nns_<mount point>
        setattr(new_cls, f"nns_{mount_point}", mount)
        # Set the sub-module/envoy to the normal mount point on the CLASS itself not the instance.
        setattr(new_cls, mount_point, envoy)

        # Update the class on the instance
        self.__class__ = new_cls

    def _set_tracer(self, tracer: Tracer, propagate=True):
        """Set tracer object on Envoy.

        Args:
            tracer (Tracer): Tracer to set.
            propagate (bool, optional): If to propagate to all sub-modules. Defaults to True.
        """

        self._tracer = tracer

        if propagate:
            for envoy in self._sub_envoys:
                envoy._set_tracer(tracer, propagate=True)

    def _scanning(self) -> bool:
        """Whether or not in scanning mode. Checks the current Tracer's Invoker.

        Returns:
            bool: _description_
        """

        try:

            return self._tracer.invoker.scanning

        except:

            return False

    def _reset_proxies(self, propagate: bool = True) -> None:
        """Sets proxies to None.

        Args:
            propagate (bool, optional): If to propagate to all sub-modules. Defaults to True.
        """

        self._output: InterventionProxy = None
        self._input: InterventionProxy = None

        if propagate:
            for envoy in self._sub_envoys:
                envoy._reset_proxies(propagate=True)

    def _reset(self, propagate: bool = True) -> None:
        """Sets _call_iter to zero. Calls ._reset_proxies as well.

        Args:
            propagate (bool, optional): If to propagate to all sub-modules. Defaults to True.
        """

        self._reset_proxies(propagate=False)

        self._call_iter = 0

        if propagate:
            for envoy in self._sub_envoys:
                envoy._reset(propagate=True)

    def _clear(self, propagate: bool = True) -> None:
        """Clears _fake_outputs and _fake_inputs. Calls ._reset as well.

        Args:
            propagate (bool, optional): If to propagate to all sub-modules. Defaults to True.
        """

        self._reset(propagate=False)

        self._fake_outputs = []
        self._fake_inputs = []

        if propagate:
            for envoy in self._sub_envoys:
                envoy._clear(propagate=True)

    def _hook(
        self, module: torch.nn.Module, input: Any, input_kwargs: Dict, output: Any
    ):

        if self._scanning():

            self._reset_proxies(propagate=False)

            input = (input, input_kwargs)

            self._fake_outputs.append(output)
            self._fake_inputs.append(input)

    def next(self, increment: int = 1, propagate: bool = False) -> Envoy:

        self._call_iter += increment

        self._reset_proxies(propagate=False)

        if propagate:
            for envoy in self._sub_envoys:
                envoy.next(increment=increment, propagate=True)

        return self

    def to(self, *args, **kwargs) -> Envoy:
        """Override torch.nn.Module.to so this returns the Envoy, not the underlying module when doing: model = model.to(...)

        Returns:
            Envoy: Envoy.
        """

        self._module = self._module.to(*args, **kwargs)

        return self

    def modules(
        self, include_fn: Callable[[Envoy], bool] = None, names: bool = False, envoys: List = None
    ) -> List[Envoy]:
        """Returns all Envoys in the Envoy tree.

        Args:
            include_fn (Callable, optional): Optional function to be ran against all Envoys to check if they should be included in the final collection of Envoys. Defaults to None.
            names (bool, optional): If to include the name/module_path of returned Envoys along with the Envoy itself. Defaults to False.

        Returns:
            List[Envoy]: Included Envoys
        """

        if envoys is None:
            envoys = list()

        included = True

        if include_fn is not None:
            included = include_fn(self)

        if included:
            if names:
                envoys.append((self.path, self))
            else:
                envoys.append(self)

        for sub_envoy in self._sub_envoys:
            sub_envoy.modules(include_fn=include_fn, names=names, envoys=envoys)

        return envoys

    def named_modules(self, *args, **kwargs) -> List[Tuple[str, Envoy]]:
        """Returns all Envoys in the Envoy tree along with their name/module_path.

        Args:
            include_fn (Callable, optional): Optional function to be ran against all Envoys to check if they should be included in the final collection of Envoys. Defaults to None.

        Returns:
            List[Tuple[str, Envoy]]: Included Envoys and their names/module_paths.
        """

        return self.modules(*args, **kwargs, names=True)

    def _repr_module_list(self):

        list_of_reprs = [repr(item) for item in self._sub_envoys]
        if len(list_of_reprs) == 0:
            return self._module._get_name() + "()"

        start_end_indices = [[0, 0]]
        repeated_blocks = [list_of_reprs[0]]
        for i, r in enumerate(list_of_reprs[1:], 1):
            if r == repeated_blocks[-1]:
                start_end_indices[-1][1] += 1
                continue

            start_end_indices.append([i, i])
            repeated_blocks.append(r)

        lines = []
        main_str = self._module._get_name() + "("
        for (start_id, end_id), b in zip(start_end_indices, repeated_blocks):
            local_repr = f"({start_id}): {b}"  # default repr

            if start_id != end_id:
                n = end_id - start_id + 1
                local_repr = f"({start_id}-{end_id}): {n} x {b}"

            local_repr = torch.nn.modules.module._addindent(local_repr, 2)
            lines.append(local_repr)

        main_str += "\n  " + "\n  ".join(lines) + "\n"
        main_str += ")"
        return main_str

    def __repr__(self) -> str:
        """Wrapper method for underlying module's string representation.

        Returns:
            str: String.
        """

        if isinstance(self._module, torch.nn.ModuleList):

            return self._repr_module_list()

        extra_lines = []
        extra_repr = self._module.extra_repr()
        # empty string will be split into list ['']
        if extra_repr:
            extra_lines = extra_repr.split("\n")
        child_lines = []
        for attribute_name, attribute in self.__dict__.items():

            if attribute_name == "_tracer":
                continue

            if isinstance(attribute, Envoy):

                mod_str = repr(attribute)
                mod_str = torch.nn.modules.module._addindent(mod_str, 2)
                child_lines.append("(" + attribute_name + "): " + mod_str)

        lines = extra_lines + child_lines

        main_str = self._module._get_name() + "("
        if lines:
            # simple one-liner info, which most builtin Modules will use
            if len(extra_lines) == 1 and not child_lines:
                main_str += extra_lines[0]
            else:
                main_str += "\n  " + "\n  ".join(lines) + "\n"

        main_str += ")"

        return main_str

    def __iter__(self) -> Iterator[Envoy]:
        """Wrapper method for underlying ModuleList iterator.

        Returns:
            Iterator[Envoy]: Iterator.
        """

        return iter(self._sub_envoys)

    def __getitem__(self, key: int) -> Envoy:
        """Wrapper method for underlying ModuleList getitem.

        Args:
            key (int): Key.

        Returns:
            Envoy: Envoy.
        """

        return self._sub_envoys[key]

    def __len__(self) -> int:
        """Wrapper method for underlying ModuleList len.

        Returns:
            int: Length.
        """

        return len(self._module)

    def __getattr__(self, key: str) -> Union[Envoy, Any]:
        """Wrapper method for underlying module's attributes.

        Args:
            key (str): Key.

        Returns:
            Any: Attribute.
        """

        return getattr(self._module, key)

    def __setattr__(self, key: Any, value: Any) -> None:
        """Overload setattr to create and set an Envoy when trying to set a torch Module."""

        if key != "_module" and isinstance(value, torch.nn.Module):

            setattr(self._module, key, value)

            self._add_envoy(value, key)

        else:

            super().__setattr__(key, value)

    def __call__(
        self, *args: List[Any], hook=False, **kwargs: Dict[str, Any]
    ) -> InterventionProxy:
        """Creates a proxy to call the underlying module's forward method with some inputs.

        Returns:
            InterventionProxy: Module call proxy.
        """

        if isinstance(self._tracer.backend, EditBackend):
            hook = True

        return protocols.ApplyModuleProtocol.add(
            self._tracer.graph, self.path, *args, hook=hook, **kwargs
        )

    @property
    def output(self) -> InterventionProxy:
        """
        Calling denotes the user wishes to get the output of the underlying module and therefore we create a Proxy of that request.
        Only generates a proxy the first time it is references otherwise return the already set one.

        Returns:
            InterventionProxy: Output proxy.
        """
        if self._output is None:

            if isinstance(self._module, torch.nn.ModuleList):

                self._output = [envoy.output for envoy in self._sub_envoys]

                return self._output

            if len(self._fake_outputs) == 0:
                fake_output = inspect._empty
            elif self._call_iter >= len(self._fake_outputs):
                # TODO warning?
                fake_output = self._fake_outputs[-1]
            else:
                fake_output = self._fake_outputs[self._call_iter]

            module_path = f"{self.path}.output"

            self._output = InterventionProtocol.add(
                self._tracer.graph,
                fake_output,
                args=[
                    module_path,
                    len(self._tracer._invoker_inputs) - 1,
                    self._call_iter,
                ],
            )

        return self._output

    @output.setter
    def output(self, value: Union[InterventionProxy, Any]) -> None:
        """
        Calling denotes the user wishes to set the output of the underlying module and therefore we create a Proxy of that request.

        Args:
            value (Union[InterventionProxy, Any]): Value to set output to.
        """

        protocols.SwapProtocol.add(self.output.node, value)

        self._output = None

    @property
    def inputs(self) -> InterventionProxy:
        """
        Calling denotes the user wishes to get the input of the underlying module and therefore we create a Proxy of that request.
        Only generates a proxy the first time it is references otherwise return the already set one.

        Returns:
            InterventionProxy: Input proxy.
        """
        if self._input is None:

            if isinstance(self._module, torch.nn.ModuleList):

                self._input = [envoy.input for envoy in self._sub_envoys]

                return self._input

            if len(self._fake_inputs) == 0:
                fake_input = inspect._empty
            elif self._call_iter >= len(self._fake_inputs):
                # TODO warning?
                fake_input = self._fake_inputs[-1]
            else:
                fake_input = self._fake_inputs[self._call_iter]

            module_path = f"{self.path}.input"

            self._input = InterventionProtocol.add(
                self._tracer.graph,
                fake_input,
                args=[
                    module_path,
                    len(self._tracer._invoker_inputs) - 1,
                    self._call_iter,
                ],
            )

        return self._input

    @inputs.setter
    def inputs(self, value: Union[InterventionProxy, Any]) -> None:
        """
        Calling denotes the user wishes to set the input of the underlying module and therefore we create a Proxy of that request.

        Args:
            value (Union[InterventionProxy, Any]): Value to set input to.
        """

        protocols.SwapProtocol.add(self.inputs.node, value)

        self._input = None

    @property
    def input(self) -> InterventionProxy:
        """Getting the first positional argument input of the model's module.

        Returns:
            InterventionProxy: Input proxy.
        """

        return self.inputs[0][0]
    
    @input.setter
    def input(self, value: Union[InterventionProxy, Any]) -> None:
        """Setting the value of the input's first positionl argument in the model's module.
        
        Args;
            value (Union[InterventionProxy, Any]): Value to set the input to.
        """
        
        self.inputs = ((value,) + self.inputs[0][1:],) + (self.inputs[1:])


"""The patching module handles patching of classes and functions in modules."""
from __future__ import annotations

import importlib
import types
from contextlib import AbstractContextManager
from typing import Any, List, Optional

from . import util


class Patch:
    """Class representing a replacement of an attribute on a module.

    Attributes:
        obj (Any): Object to replace.
        replacement (Any): Object that replaces.
        parent (Any): Module or class to replace attribute.
    """

    def __init__(self, parent: Any, replacement: Any, key: str) -> None:
        self.parent = parent
        self.replacement = replacement
        self.key = key
        self.orig = getattr(self.parent, key)

    def patch(self) -> None:
        """Carries out the replacement of an object in a module/class."""
        setattr(self.parent, self.key, self.replacement)

    def restore(self) -> None:
        """Carries out the restoration of the original object on the objects module/class."""

        setattr(self.parent, self.key, self.orig)


class Patcher(AbstractContextManager):
    """Context manager that patches from a list of Patches on __enter__ and restores the patch on __exit__.

    Attributes:
        patches (List[Patch]):
    """

    def __init__(self, patches: Optional[List[Patch]] = None) -> None:
        self.patches = patches or []

    def add(self, patch: Patch) -> None:
        """Adds a Patch to the patches. Also calls `.patch()` on the Patch.

        Args:
            patch (Patch): Patch to add.
        """
        self.patches.append(patch)

        patch.patch()

    def __enter__(self) -> Patcher:
        """Enters the patching context. Calls `.patch()` on all patches.

        Returns:
            Patcher: Patcher
        """
        for patch in self.patches:
            patch.patch()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Calls `.restore()` on all patches."""
        for patch in self.patches:
            patch.restore()



"""This module contains the main NNsight model classes which enable the tracing and interleaving functionality of nnsight. 

Models allow users to load and wrap torch modules. Here we load gpt2 from HuggingFace using its repo id:

.. code-block:: python

    from nnsight import LanguageModel
    model = LanguageModel('gpt2', device_map='cuda:0')

In this case, declaring a LanguageModel entails the underlying model is a ``transformers.AutoModelForCausalLM``, and unused arguments by LanguageModel are passed downstream to AutoModelForCausalLM. ``device_map='cuda:0'`` leverages the accelerate package to use the first GPU when loading the local model.

The wrapping of the underlying model encompasses both the display of its module structure when the model is printed and the ability to directly access the attributes of the underlying meta_model through the wrapped model itself.

Printing out the wrapped module returns its structure:

.. code-block:: python

    print(model)

.. code-block:: text

    GPT2LMHeadModel(
      (transformer): GPT2Model(
        (wte): Embedding(50257, 768)
        (wpe): Embedding(1024, 768)
        (drop): Dropout(p=0.1, inplace=False)
        (h): ModuleList(
          (0-11): 12 x GPT2Block(
            (ln_1): LayerNorm((768,), eps=1e-05, elementwise_affine=True)
            (attn): GPT2AttentionAltered(
              (c_attn): Conv1D()
              (c_proj): Conv1D()
              (attn_dropout): Dropout(p=0.1, inplace=False)
              (resid_dropout): Dropout(p=0.1, inplace=False)
              (query): WrapperModule()
              (key): WrapperModule()
              (value): WrapperModule()
            )
            (ln_2): LayerNorm((768,), eps=1e-05, elementwise_affine=True)
            (mlp): GPT2MLP(
              (c_fc): Conv1D()
              (c_proj): Conv1D()
              (act): NewGELUActivation()
              (dropout): Dropout(p=0.1, inplace=False)
            )
          )
        )
        (ln_f): LayerNorm((768,), eps=1e-05, elementwise_affine=True)
      )
      (lm_head): Linear(in_features=768, out_features=50257, bias=False)
    )

The primary method of interacting and running the model is ``.trace(...)``. This returns a context manager object which, when entered, track operations performed on the inputs and outputs of modules.



The :func:`trace <nnsight.models.NNsightModel.NNsightModel.trace>` context is has the most explicit control of all levels of nnsight tracing and interleaving, creating a parent context where sub, input specific, contexts are spawned from.

.. code-block:: python

    with model.trace("The Eiffel Tower is in the city of") as runner:
        logits = model.lm_head.output.save()

    print(logits.value)

See :mod:`nnsight.contexts` for more.

"""

from .NNsightModel import NNsight



from __future__ import annotations

import inspect
import weakref
from collections import defaultdict
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union

import torch

from .. import util
from ..logger import logger
from . import protocols
from .Proxy import Proxy

if TYPE_CHECKING:
    from .Graph import Graph

    try:
        from pygraphviz import AGraph
    except:
        pass


class Node:
    """A Node represents some action that should be carried out during execution of a Graph.

    The class represents the operations (and the resulting output of said operations) they are tracing AND nodes that actually execute the operations when executing the Graph. The Nodes you are Tracing are the same object as the ones that are executed.

        * Nodes have a ``.proxy_value`` attribute that are a result of the tracing operation, and are FakeTensors allowing you to view the shape and datatypes of the actual resulting value that will be populated when the node' operation is executed.
        * Nodes carry out their operation in ``.execute()`` where their arguments are pre-processed and their value is set in ``.set_value()``.
        * Arguments passed to the node are other nodes, where a bi-directional dependency graph is formed. During execution pre-processing, the arguments that are nodes and converted to their value.
        * Nodes are responsible for updating their listeners that one of their dependencies are completed, and if all are completed that they should execute. Similarly, nodes must inform their dependencies when one of their listeners has ceased "listening." If the node has no listeners, it's value is destroyed by calling ``.destroy()`` in order to free memory. When re-executing the same graph and therefore the same nodes, the remaining listeners and dependencies are reset on each node.

    Attributes:
        name (str): Unique name of node.
        graph (Graph): Weak reference to parent Graph object.
        proxy (Proxy): Weak reference to Proxy created from this Node.
        proxy_value (Any): Fake Tensor version of value. Used when graph has validate = True to check of Node actions are possible.
        target (Union[Callable, str]): Function to execute or name of Protocol.
        args (List[Any], optional): Positional arguments. Defaults to None.
        kwargs (Dict[str, Any], optional): Keyword arguments. Defaults to None.
        listeners (List[Node]): Nodes that depend on this node.
        arg_dependencies (List[Node]): Nodes that this node depends on.
        cond_dependency (Optional[Node]): ConditionalProtocol node if this node was defined within a Conditional context.
        value (Any): Actual value to be populated during execution.
    """

    def __init__(
        self,
        target: Union[Callable, str],
        graph: "Graph" = None,
        proxy_value: Any = inspect._empty,
        args: List[Any] = None,
        kwargs: Dict[str, Any] = None,
        name: str = None,
    ) -> None:
        super().__init__()

        if args is None:
            args = list()
        if kwargs is None:
            kwargs = dict()

        args = list(args)

        self.graph: "Graph" = graph
        self.proxy_value = proxy_value
        self.target = target
        self.args, self.kwargs = args, kwargs

        self.proxy: Optional[Proxy] = None

        self._value: Any = inspect._empty

        self.listeners: List[Node] = list()
        self.arg_dependencies: List[Node] = list()
        self.cond_dependency: Optional[Node] = None

        self.remaining_listeners = 0
        self.remaining_dependencies = 0

        # Preprocess args.
        self.preprocess()

        # Node.graph is a weak reference to avoid reference loops.
        self.graph = (
            weakref.proxy(self.graph) if self.graph is not None else None
        )

        self.name: str = name

        # If theres an alive Graph, add it.
        if self.attached():

            self.graph.add(self)

    def preprocess(self) -> None:
        """Preprocess Node.args and Node.kwargs."""

        # bridge graph redirection
        if self.attached():
            self.graph = (
                protocols.BridgeProtocol.peek_graph(self.graph)
                if (
                    self.target.redirect
                    if isinstance(self.target, type)
                    and issubclass(self.target, protocols.Protocol)
                    else True
                )
                else self.graph
            )

        def preprocess_node(node: Union[Node, Proxy]):

            if isinstance(node, Proxy):

                node = node.node

            if node.done():

                return node.value

            if self.attached() and self.graph.id != node.graph.id:

                node = protocols.BridgeProtocol.add(node).node

            self.arg_dependencies.append(node)
            # Weakref so no reference loop
            node.listeners.append(weakref.proxy(self))

            return node

        self.args, self.kwargs = util.apply(
            (self.args, self.kwargs), preprocess_node, (Node, Proxy)
        )

        # conditional context handling
        if (
            self.attached()
            and protocols.ConditionalProtocol.has_conditional(self.graph)
            and (
                self.target.condition
                if isinstance(self.target, type)
                and issubclass(self.target, protocols.Protocol)
                else True
            )
        ):

            conditional_node = protocols.ConditionalProtocol.peek_conditional(
                self.graph
            )

            # only the top dependency needs to add the Conditional as a dependency
            # if none of the dependent are dependent on the Conditional, then add it
            if conditional_node:
                if all(
                    [
                        not protocols.ConditionalProtocol.is_node_conditioned(
                            arg
                        )
                        for arg in self.arg_dependencies
                    ]
                ):
                    self.cond_dependency = conditional_node
                    conditional_node.listeners.append(weakref.proxy(self))

                protocols.ConditionalProtocol.add_conditioned_node(self)

    @property
    def value(self) -> Any:
        """Property to return the value of this node.

        Returns:
            Any: The stored value of the node, populated during execution of the model.

        Raises:
            ValueError: If the underlying ._value is inspect._empty (therefore never set or destroyed).
        """

        if not self.done():
            raise ValueError("Accessing value before it's been set.")

        return self._value

    def attached(self) -> bool:
        """Checks to see if the weakref to the Graph is alive or dead.

        Returns:
            bool: Is Node attached.
        """

        try:

            return self.graph.alive

        except:
            return False

    def create(
        self,
        target: Union[Callable, str],
        proxy_value: Any = inspect._empty,
        args: List[Any] = None,
        kwargs: Dict[str, Any] = None,
        name: str = None,
    ) -> Union[Proxy, Any]:
        """We use Node.add vs Graph.add in case graph is dead.
        If the graph is dead, we assume this node is ready to execute and therefore we try and execute it and then return its value.

        Returns:
            Union[Proxy, Any]: Proxy or value
        """

        if not self.attached():

            graph: "Graph" = None

            def find_attached_graph(node: Union[Proxy, Node]):

                if isinstance(node, Proxy):

                    node = node.node

                nonlocal graph

                if node.attached():

                    graph = node.graph

            util.apply((args, kwargs), find_attached_graph, (Proxy, Node))

            if graph is not None:

                return graph.create(
                    target=target,
                    name=name,
                    proxy_value=proxy_value,
                    args=args,
                    kwargs=kwargs,
                )

            # Create Node with no values or Graph.
            node = Node(
                target=target,
                graph=None,
                proxy_value=None,
                args=args,
                kwargs=kwargs,
            )

            # Reset it.
            node.reset()

            # So it doesn't get destroyed.
            node.remaining_listeners = 1

            # Execute Node
            node.execute()

            # Get value.
            value = node.value

            # Destroy.
            node.destroy()

            return value

        # Otherwise just create the Node on the Graph like normal.
        return self.graph.create(
            target=target,
            name=name,
            proxy_value=proxy_value,
            args=args,
            kwargs=kwargs,
        )

    def reset(self) -> None:
        """Resets this Nodes remaining_listeners and remaining_dependencies."""

        self.remaining_listeners = len(self.listeners)
        self.remaining_dependencies = len(self.arg_dependencies) + int(
            not (self.cond_dependency is None)
        )

    def done(self) -> bool:
        """Returns true if the value of this node has been set.

        Returns:
            bool: If done.
        """
        return self._value is not inspect._empty

    def executed(self) -> bool:
        """Returns true if remaining_dependencies is less than 0.

        Returns:
            bool: If executed.
        """
        return self.remaining_dependencies < 0

    def fulfilled(self) -> bool:
        """Returns true if remaining_dependencies is 0.

        Returns:
            bool: If fulfilled.
        """
        return self.remaining_dependencies == 0

    def redundant(self) -> bool:
        """Returns true if remaining_listeners is 0.

        Returns:
            bool: If redundant.
        """
        return self.remaining_listeners == 0

    @classmethod
    def prepare_inputs(
        cls, inputs: Any, device: torch.device = None, proxy: bool = False
    ) -> Any:
        """Prepare arguments for executing this node's target.
        Converts Nodes in args and kwargs to their value and moves tensors to correct device.

        Returns:
            Any: Prepared inputs.
        """

        inputs = util.apply(inputs, lambda x: x, inspect._empty)

        def _value(node: Proxy | Node):

            if isinstance(node, Proxy):
                node = node.node

            if proxy:
                return node.proxy_value

            return node.value

        inputs = util.apply(inputs, _value, (Node, Proxy), inplace=not proxy)

        if device is None:

            def _device(value: torch.Tensor):
                nonlocal device

                if device is None:
                    device = value.device

            util.apply(inputs, _device, torch.Tensor)

        def _to(value: torch.Tensor):
            return value.to(device)

        inputs = util.apply(inputs, _to, torch.Tensor, inplace=not proxy)

        return inputs

    def execute(self) -> None:
        """Actually executes this node.
        Lets protocol execute if target is str.
        Else prepares args and kwargs and passes them to target. Gets output of target and sets the Node's value to it.
        """

        try:

            if isinstance(self.target, type) and issubclass(
                self.target, protocols.Protocol
            ):

                self.target.execute(self)

            else:

                # Prepare arguments.
                args, kwargs = Node.prepare_inputs((self.args, self.kwargs))

                # Call the target to get value.
                output = self.target(*args, **kwargs)

                # Set value.
                self.set_value(output)

        except Exception as e:

            raise type(e)(
                f"Above exception when execution Node: '{self.name}' in Graph: '{self.graph.id}'"
            ) from e

        finally:
            self.remaining_dependencies -= 1

    def set_value(self, value: Any) -> None:
        """Sets the value of this Node and logs the event.
        Updates remaining_dependencies of listeners. If they are now fulfilled, execute them.
        Updates remaining_listeners of dependencies. If they are now redundant, destroy them.

        Args:
            value (Any): Value.
        """
        self._value = value

        logger.info(f"=> SET({self.name})")

        for listener in self.listeners:
            listener.remaining_dependencies -= 1

            if listener.fulfilled() and not self.graph.sequential:
                listener.execute()

        for dependency in self.arg_dependencies:
            dependency.remaining_listeners -= 1

            if dependency.redundant():
                dependency.destroy()

        if self.done() and self.redundant():
            self.destroy()

    def destroy(self) -> None:
        """Removes the reference to the node's value and logs it's destruction."""

        logger.info(f"=> DEL({self.name})")

        self._value = inspect._empty

    def clean(self) -> None:
        """Clean up dependencies during early execution stop"""

        # BridgeProtocol nodes must clean up their corresponding external proxy
        if isinstance(self.target, type) and issubclass(
            self.target, protocols.BridgeProtocol
        ):
            bridge = protocols.BridgeProtocol.get_bridge(self.graph)
            lock_node = bridge.get_graph(self.args[0]).nodes[self.args[1]]
            lock_dependency = lock_node.args[0]
            lock_dependency.remaining_listeners -= 1
            lock_node.destroy()
            if lock_dependency.redundant():
                lock_dependency.destroy()
        else:
            for dependency in self.arg_dependencies:
                dependency.remaining_listeners -= 1
                if dependency.redundant():
                    dependency.destroy()

    def visualize(
        self, viz_graph: "AGraph", recursive: bool, backend_name: str = ""
    ) -> str:
        """Adds this node to the visualization graph and recursively visualizes its arguments and adds edges between them.

        Args:
            - viz_graph (AGraph): Visualization graph.
            - recursive (bool): If True, recursively visualizes all sub-graphs.
            - backend_name (str): Inherent parent graph name for unique differentiation in recursive visualization.

        Returns:
            - str: name of this node.
        """

        styles = {
            "node": {"color": "black", "shape": "ellipse"},
            "label": (self.target if isinstance(self.target, str) else self.target.__name__),
            "arg": defaultdict(lambda: {"color": "gray", "shape": "box"}),
            "arg_kname": defaultdict(lambda: None),
            "edge": defaultdict(lambda: "solid"),
        }

        node_name = backend_name + self.name

        if isinstance(self.target, type) and issubclass(
            self.target, protocols.Protocol
        ):
            styles = self.target.style()
            viz_graph.add_node(node_name, label=styles["label"], **styles["node"])
            if (
                recursive
                and self.target == protocols.LocalBackendExecuteProtocol
            ):
                # recursively draw all sub-graphs
                for sub_node in self.args[0].graph.nodes.values():
                    # draw root nodes and attach them to their LocalBackendExecuteProtocol node
                    if (
                        len(sub_node.arg_dependencies)
                        + int(not (sub_node.cond_dependency is None))
                    ) == 0:
                        sub_node_name = sub_node.visualize(
                            viz_graph, recursive, node_name + "_"
                        )
                        viz_graph.add_edge(
                            node_name,
                            sub_node_name,
                            style="dotted",
                            color="purple",
                        )
                    # draw bottom up
                    elif len(sub_node.listeners) == 0:
                        sub_node_name = sub_node.visualize(
                            viz_graph, recursive, node_name + "_"
                        )
        else:
            viz_graph.add_node(node_name, label=styles["label"], **styles["node"])

        def visualize_args(arg_collection):
            """Recursively visualizes the arguments of this node.

            Args:
                - arg_collection (Union[List[Any], Dict[str, Any]]): Collection of Node arguments.
            """

            for key, arg in arg_collection:
                if isinstance(arg, Node):
                    name = arg.visualize(viz_graph, recursive, backend_name)
                else:
                    # show link between iterable values with Node dependencies
                    iter_val_dependencies = []
                    if isinstance(arg, Iterable):
                        for element in arg:
                            if isinstance(element, Node):
                                dep_name = element.visualize(viz_graph, recursive, backend_name)
                                iter_val_dependencies.append(dep_name)
                    
                    name = node_name
                    if isinstance(arg, torch.Tensor):
                        name += f"_Tensor_{key}"
                        label = "Tensor"
                    elif isinstance(arg, str):
                        name += f"_{arg}_{key}"
                        label = f'"{arg}"'
                    else:
                        name += f"_{arg}_{key}"
                        label = str(arg)

                    if isinstance(key, int):
                        if not styles["arg_kname"][key] is None:
                            label = f"{styles['arg_kname'][key]}={label}"
                    else:
                        label = f"{key}={label}"

                    viz_graph.add_node(name, label=label, **styles["arg"][key])

                    for dep_name in iter_val_dependencies:
                        viz_graph.add_edge(dep_name, name, style="dashed", color="gray")

                viz_graph.add_edge(name, node_name, style=styles["edge"][key])

        visualize_args(enumerate(self.args))

        visualize_args(self.kwargs.items())

        if isinstance(self.cond_dependency, Node):
            name = self.cond_dependency.visualize(
                viz_graph, recursive, backend_name
            )
            viz_graph.add_edge(
                name, node_name, style=styles["edge"][None], color="#FF8C00"
            )

        return node_name

    def __str__(self) -> str:
        args = util.apply(self.args, lambda x: f"'{x}'", str)
        args = util.apply(args, lambda x: x.name, Node)
        args = [str(arg) for arg in args]

        return f"{self.name}:[args:({','.join(args)}) l:{len(self.listeners)} a_d:{len(self.arg_dependencies)} c_d{bool(self.cond_dependency)}]"

    def __repr__(self) -> str:
        return f"&lt;{self.__class__.__name__} at {hex(id(self))}&gt;"



from __future__ import annotations

import logging
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional, Union

import torch
from pydantic import BaseModel

from .. import util
from ..tracing.Graph import Graph


class ResultModel(BaseModel):
    id: str
    value: Any = None

    @classmethod
    def from_graph(cls, graph: Graph) -> Dict[str, Any]:

        saves = {
            name: util.apply(node.value, lambda x: x.detach().cpu(), torch.Tensor)
            for name, node in graph.nodes.items()
            if node.done()
        }

        return saves
 

class ResponseModel(BaseModel):
    class JobStatus(Enum):
        RECEIVED = "RECEIVED"
        APPROVED = "APPROVED"
        RUNNING = "RUNNING"
        COMPLETED = "COMPLETED"
        LOG = "LOG"
        ERROR = "ERROR"

    id: str
    status: JobStatus
    description: str

    received: datetime = None
    session_id: Optional[str] = None

    def __str__(self) -> str:
        return f"{self.id} - {self.status.name}: {self.description}"

    def log(self, logger: logging.Logger) -> ResponseModel:
        if self.status == ResponseModel.JobStatus.ERROR:
            logger.error(str(self))
        else:
            logger.info(str(self))

        return self



import torch
from abc import ABC, abstractmethod

from nnsight import LanguageModel, Module
from typing import Any, Callable, Dict, List, Union

from functools import reduce

class Lens(ABC):

    def __init__(self):
        super().__init__()
        self.tuned = False
        
    @abstractmethod
    def __call__(self) -> Any:
        pass

class LogitLens(Lens):
    """Returns the probability distribution over all tokens at specified points in the model.
    
    """

    def __init__(self, 
                 layers: List[Module],
                 decoding_modules: List[Module],
                ) -> None:
        super().__init__()
        
        self.tuned = False
        
        self.layers = layers
        self.decoder = lambda x: reduce(lambda acc, func: func(acc), decoding_modules, x)

    def __call__(
            self,
            indices: Union[int, List] = None,
            as_probs: bool = True,
        ) -> List[Any]:
            
        observations = []
        
        for layer in self.layers:
            logits = self.decoder(layer.output[0]) # apply decoder to hidden state

            observations.append(logits.save())

        # Return logits over a specific token
        if type(indices) == List or type(indices == int):
            observations = [logits[:,indices,:] for logits in observations]
            
        # Raw logits to probabilities
        if as_probs:
            observations = [logits.softmax(dim=-1) for logits in observations]

        self.observations = observations

class TunedLens(Lens):

    def __init__(self):
        super().__init__()
        self.tuned = True




"""The `nnsight.tracing `module involves tracing operations in order to form a computation graph.

The :class:`Graph <nnsight.tracing.Graph.Graph>` class adds and stores operations as `Node`s . 

The :class:`Node <nnsight.tracing.Node.Node>` class represents an individual operation in the :class:`Graph <nnsight.tracing.Graph.Graph>`.

The :class:`Proxy <nnsight.tracing.Proxy.Proxy>` class handles interactions from the user in order to create new `Node`s. There is a `Proxy` for each `Node`.
"""



"""The contexts module contains logic for managing the tracing and running of models with :mod:`nnsight.tracing` and :mod:`nnsight.envoy`

The primary two classes involved here are :class:`Tracer <nnsight.contexts.Tracer.Tracer>` and :class:`Invoker <nnsight.contexts.Invoker.Invoker>`.

The :class:`Tracer <nnsight.contexts.Tracer.Tracer>` class creates a :class:`Graph <nnsight.tracing.Graph.Graph>` around the underlying model of an :class:`NNsight <nnsight.models.NNsightModel.NNsight>`. The graph tracks and manages the operations performed on the inputs and outputs of said model.
Module's envoys in the model expose their ``.output`` and ``.input`` attributes which when accessed, add to the computation graph of the tracer.
To do this, they need to know about the current Tracer object, so each module's envoy's ``.tracer`` object is set to be the current Tracer.

The Tracer object also keeps track of the batch_size of the most recent input, the generation index for multi iteration generations, and all of the inputs made during its context in the ``.batched_input`` attribute. Inputs added to this attribute should be in a format where each index is a batch index and allows the model to batch all of the inputs together.

This is to keep things consistent. If two different inputs are in two different valid formats, they both become the same format and are easy to batch.
In the case of LanguageModels, regardless of whether the input are string prompts, pre-processed dictionaries, or input ids, the batched input is only input ids.
On exiting the Tracer context, the Tracer object should use the information and inputs provided to it to carry out the execution of the model.

The :class:`Invoker <nnsight.contexts.Invoker.Invoker>` class is what actually accepts inputs to the model/graph, and it updates its parent Tracer object with the appropriate information about respective inputs. On entering the invoker context with some input, the invoker leverages the model to pre-process and prepare the input to the model. Using the prepared inputs, it updates its Tracer object with a batched version of the input, the size of the batched input, and the current generation index. It also runs a 'meta' version of the input through the model's meta_model. This updates the sizes/dtypes of all of the module's Envoys inputs and outputs based on the characteristics of the input.

nnsight comes with an extension of a Tracer, RemoteTracer,  which enables both local and remote execution.
"""

from typing import Any, Tuple

from .. import util
from ..tracing import protocols
from ..tracing.Node import Node
from ..tracing.Proxy import Proxy


def check_for_dependencies(data: Any) -> Tuple[Any, bool]:
    """Checks to see if there are any Proxies in data.
    If so, convert them to a Bridge Node, then a Lock Node in order to
    later get the value of the Bridge Node come execution.

    Args:
        data (Any): Data to check for Proxies.

    Returns:
        Any: Data with Proxies replaced with Bridge/Lock Nodes.
        bool: If there were any proxies in data.
    """

    has_proxies = False

    def check_for_nodes(proxy: Proxy):

        if not proxy.node.done():

            nonlocal has_proxies

            has_proxies = True

            node = proxy.node

            return protocols.LockProtocol.add(
                protocols.BridgeProtocol.add(
                    node,
                ).node
            ).node

        else:

            return proxy.node.value

    return util.apply(data, check_for_nodes, Proxy), has_proxies


def resolve_dependencies(data: Any) -> Any:
    """Turn any dependencies (Locked Bridge Node) within data into their value.
    Executes the Bridge Node.

    Args:
        data (Any): Data to find and resolve dependencies within.

    Returns:
        Any: Data with dependencies converted to their value.
    """

    def get_value(node: Node):
        
        bridge_node: Node = node.args[0]

        bridge_node.execute()
            
        # Get value of Bridge Node
        value = bridge_node.value

        # Clear Lock Node
        node.set_value(None)

        return value

    return util.apply(data, get_value, Node)



from typing import TYPE_CHECKING, Any, Callable, List, Tuple, Union

from . import Backend

if TYPE_CHECKING:
    from ...tracing.Bridge import Bridge


class BridgeMixin:
    """To be inherited by objects that want to be able to be executed by the BridgeBackend."""

    def bridge_backend_handle(self, bridge: "Bridge") -> None:
        """Should add self to the current Bridge in some capacity.

        Args:
            bridge (Bridge): Current Bridge.
        """

        raise NotImplementedError()


class BridgeBackend(Backend):
    """Backend to accumulate multiple context object to be executed collectively.

    Context object must inherit from BridgeMixin and implement its methods.

    Attributes:

        bridge (Bridge): Current Bridge object.
    """

    def __init__(self, bridge: "Bridge") -> None:

        self.bridge = bridge

    def __call__(self, obj: BridgeMixin):

        obj.bridge_backend_handle(self.bridge)



from __future__ import annotations

import weakref
from collections.abc import Iterable
from typing import TYPE_CHECKING, Iterable, Tuple, Union

from ... import util
from ...tracing import protocols
from ...tracing.Node import Node
from .. import check_for_dependencies, resolve_dependencies
from ..GraphBasedContext import GraphBasedContext

if TYPE_CHECKING:
    from ...intervention import InterventionProxy
    from ...tracing.Bridge import Bridge


class Iterator(GraphBasedContext):
    """Intervention loop context for iterative execution of an intervention graph.

    Attributes:
        - data (Iterable): Data to iterate over.
        - return_context (bool): If True, returns the Iterator object upon entering the Iterator context.
    """

    def __init__(
        self, data: Iterable, *args, return_context: bool = False, **kwargs
    ) -> None:

        self.data: Iterable = data
        self._return_context: bool = return_context

        super().__init__(*args, **kwargs)

    def __enter__(
        self,
    ) -> Union["InterventionProxy", Tuple["InterventionProxy", Iterator]]:

        super().__enter__()

        self.data, has_dependencies = check_for_dependencies(self.data)

        proxy_value = None

        if self.graph.validate:

            proxy_value = util.apply(
                self.data, lambda node: node.args[0].proxy_value, Node
            )

            if len(proxy_value) != 0:

                proxy_value = (
                    next(proxy_value)
                    if hasattr(proxy_value, "__next__")
                    else proxy_value[0]
                )

        iter_item_proxy: "InterventionProxy" = protocols.ValueProtocol.add(
            self.graph, proxy_value
        )

        if self._return_context:
            return iter_item_proxy, self
        else:
            return iter_item_proxy

    ### BACKENDS ########

    def local_backend_execute(self) -> None:

        self.graph.reset()

        bridge: "Bridge" = protocols.BridgeProtocol.get_bridge(self.graph)

        bridge.locks += 1

        data = resolve_dependencies(self.data)

        last_idx: int = len(data) - 1

        for idx, item in enumerate(data):

            if idx != 0:

                self.graph.reset()

            last_iter = idx == last_idx

            if last_iter:

                bridge.locks -= 1

            protocols.ValueProtocol.set(
                self.graph.nodes[f"{protocols.ValueProtocol.__name__}_0"], item
            )

            try:
                self.graph.execute()
            except protocols.EarlyStopProtocol.EarlyStopException as e:
                break
            finally:
                graph = self.graph
                graph.alive = False

                if not isinstance(graph, weakref.ProxyType):
                    self.graph = weakref.proxy(graph)

    def __repr__(self) -> str:
        return f"&lt;{self.__class__.__name__} at {hex(id(self))}&gt;"



from __future__ import annotations

import io
from typing import TYPE_CHECKING, Any, Callable

import requests
import socketio
import torch
from tqdm import tqdm

from ... import CONFIG
from ...logger import logger, remote_logger
from .LocalBackend import LocalBackend, LocalMixin

if TYPE_CHECKING:

    from ...schema.Request import RequestModel
    from ...schema.Response import ResponseModel


class RemoteMixin(LocalMixin):
    """To be inherited by objects that want to be able to be executed by the RemoteBackend."""

    def remote_backend_get_model_key(self) -> str:
        """Should return the model_key used to specify which model to run on the remote service.

        Returns:
            str: Model key.
        """

        raise NotImplementedError()

    def remote_backend_postprocess_result(self, local_result: Any) -> Any:
        """Should handle postprocessing the result from local_backend_execute.

        For example moving tensors to cpu/detaching/etc.

        Args:
            local_result (Any): Local execution result.

        Returns:
            Any: Post processed local execution result.
        """

        raise NotImplementedError()

    def remote_backend_handle_result_value(self, value: Any) -> None:
        """Should handle postprocessed result from remote_backend_postprocess_result on return from remote service.

        Args:
            value (Any): Result.
        """

        raise NotImplementedError()

    def remote_backend_cleanup(self):
        raise NotImplementedError()


class RemoteBackend(LocalBackend):
    """Backend to execute a context object via a remote service.

    Context object must inherit from RemoteMixin and implement its methods.

    Attributes:

        url (str): Remote host url. Defaults to that set in CONFIG.API.HOST.
    """

    def __init__(
        self,
        host: str = None,
        blocking: bool = True,
        job_id: str = None,
        ssl: bool = None,
        api_key: str = "",
    ) -> None:

        self.job_id = job_id or CONFIG.API.JOB_ID
        self.ssl = CONFIG.API.SSL if ssl is None else ssl
        self.api_key = api_key or CONFIG.API.APIKEY
        self.blocking = blocking
        self.handle_result = None

        self.host = host or CONFIG.API.HOST
        self.address = f"http{'s' if self.ssl else ''}://{self.host}"
        self.ws_address = f"ws{'s' if CONFIG.API.SSL else ''}://{self.host}"

    def request(self, obj: RemoteMixin):

        model_key = obj.remote_backend_get_model_key()

        from ...schema.Request import RequestModel

        # Create request using pydantic to parse the object itself.
        return RequestModel(object=obj, model_key=model_key)

    def __call__(self, obj: RemoteMixin):

        self.handle_result = obj.remote_backend_handle_result_value

        if self.blocking:

            request = self.request(obj)

            # Do blocking request.
            self.blocking_request(request)

        else:

            request = None

            if not self.job_id:

                request = self.request(obj)

            self.non_blocking_request(request=request)

        obj.remote_backend_cleanup()

    def handle_response(self, data: Any) -> "ResponseModel":
        """Handles incoming response data.

        Parses it into the `ResponseModel` pydantic object.
        Logs the response object.
        If the job is completed, retrieve and stream the result from the remote endpoint.
        Use torch.load to decode and load the `ResultModel` into memory.
        Use the backend object's .handle_result method to handle the decoded result.

        Args:
            data (Any): Json data to concert to `ResponseModel`

        Raises:
            Exception: If the job's status is `ResponseModel.JobStatus.ERROR`

        Returns:
            ResponseModel: ResponseModel.
        """

        from ...schema.Response import ResponseModel, ResultModel

        # Load the data into the ResponseModel pydantic class.
        response = ResponseModel(**data)

        # Log response for user
        remote_logger.info(str(response))

        # If the status of the response is completed, update the local nodes that the user specified to save.
        # Then disconnect and continue.
        if response.status == ResponseModel.JobStatus.COMPLETED:
            # Create BytesIO object to store bytes received from server in.
            result_bytes = io.BytesIO()
            result_bytes.seek(0)

            # Get result from result url using job id.
            with requests.get(
                url=f"{self.address}/result/{response.id}",
                stream=True,
            ) as stream:
                # Total size of incoming data.
                total_size = float(stream.headers["Content-length"])

                with tqdm(
                    total=total_size,
                    unit="B",
                    unit_scale=True,
                    desc="Downloading result",
                ) as progress_bar:
                    # chunk_size=None so server determines chunk size.
                    for data in stream.iter_content(chunk_size=None):
                        progress_bar.update(len(data))
                        result_bytes.write(data)

            # Move cursor to beginning of bytes.
            result_bytes.seek(0)

            # Decode bytes with pickle and then into pydantic object.
            result: "ResultModel" = ResultModel(
                **torch.load(
                    result_bytes, map_location="cpu", weights_only=False
                )
            )

            # Close bytes
            result_bytes.close()

            # Handle result
            self.handle_result(result.value)

        # Or if there was some error.
        elif response.status == ResponseModel.JobStatus.ERROR:
            raise Exception(str(response))

        return response

    def submit_request(self, request: "RequestModel") -> "ResponseModel":
        """Sends request to the remote endpoint and handles the response object.

        Raises:
            Exception: If there was a status code other than 200 for the response.

        Returns:
            (ResponseModel): Response.
        """

        response = requests.post(
            f"{self.address}/request",
            json=request.model_dump(exclude=["id", "received"]),
            headers={"ndif-api-key": self.api_key},
        )

        if response.status_code == 200:

            return self.handle_response(response.json())

        else:

            raise Exception(response.reason)

    def get_response(self) -> "ResponseModel":
        """Retrieves and handles the response object from the remote endpoint.

        Raises:
            Exception: If there was a status code other than 200 for the response.

        Returns:
            (ResponseModel): Response.
        """

        response = requests.get(
            f"{self.address}/response/{self.job_id}",
            headers={"ndif-api-key": self.api_key},
        )

        if response.status_code == 200:

            return self.handle_response(response.json())

        else:

            raise Exception(response.reason)

    def blocking_request(self, request: "RequestModel"):
        """Send intervention request to the remote service while waiting for updates via websocket.

        Args:
            request (RequestModel): Request.
        """

        from ...schema.Response import ResponseModel

        # Create a socketio connection to the server.
        with socketio.SimpleClient(
            logger=logger, reconnection_attempts=10
        ) as sio:
            # Connect
            sio.connect(
                self.ws_address,
                socketio_path="/ws/socket.io",
                transports=["websocket"],
                wait_timeout=10,
            )

            # Give request session ID so server knows to respond via websockets to us.
            request.session_id = sio.sid

            # Submit request via
            self.submit_request(request)

            # Loop until
            while True:
                if (
                    self.handle_response(sio.receive()[1]).status
                    == ResponseModel.JobStatus.COMPLETED
                ):
                    break

    def non_blocking_request(self, request: "RequestModel" = None):
        """Send intervention request to the remote service if request provided. Otherwise get job status.

        Sets CONFIG.API.JOB_ID on initial request as to later get the status of said job.

        When job is completed, clear CONFIG.API.JOB_ID to request a new job.

        Args:
            request (RequestModel): Request if submitting a new request. Defaults to None
        """

        from ...schema.Response import ResponseModel

        if request is not None:

            # Submit request via
            response = self.submit_request(request)

            CONFIG.API.JOB_ID = response.id

            CONFIG.save()

        else:

            try:

                response = self.get_response()

                if response.status == ResponseModel.JobStatus.COMPLETED:

                    CONFIG.API.JOB_ID = None

                    CONFIG.save()

            except Exception as e:

                CONFIG.API.JOB_ID = None

                CONFIG.save()

                raise e



from typing import Any

from . import Backend


class EditMixin:
    """To be inherited by objects that want to be able to be executed by the EditBackend."""

    def edit_backend_execute(self) -> Any:
        """Should execute this object locally and return a result that can be handled by EditMixin objects.

        Returns:
            Any: Result containing data to return from a edit execution.
        """

        raise NotImplementedError()


class EditBackend(Backend):
    """Backend to execute a default edit.

    Context object must inherit from EditMixin and implement its methods.
    """

    def __call__(self, obj: EditMixin):

        obj.edit_backend_execute()



from __future__ import annotations

import gc
import inspect
import weakref
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union

import accelerate
import torch
from accelerate import init_empty_weights
from torch.utils._python_dispatch import TorchDispatchMode
from transformers import AutoConfig, AutoModel
from typing_extensions import Self

from .. import util
from ..contexts.backends import (
    Backend,
    BridgeBackend,
    EditBackend,
    LocalBackend,
    NoopBackend,
    RemoteBackend,
)
from ..contexts.session.Session import Session
from ..contexts.Tracer import Tracer
from ..envoy import Envoy
from ..intervention import (
    HookHandler,
    InterventionHandler,
    InterventionProtocol,
    InterventionProxy,
)
from ..logger import logger
from ..tracing import protocols
from ..tracing.Graph import Graph


class MetaDispatcher(TorchDispatchMode):
    """This exists because `with torch.device('meta') is evil.

    Ty Caden

    """

    def __torch_dispatch__(self, func, types, args, kwargs=None):
        if kwargs is None:
            kwargs = {}

        if "device" in kwargs or "device" in inspect.signature(func).parameters:
            kwargs["device"] = "meta"
        return func(*args, **kwargs)


class NNsight:
    """Main class to be implemented as a wrapper for PyTorch models wishing to gain this package's functionality. Can be used "as is" for basic models.

    Class Attributes:

        proxy_class (Type[InterventionProxy]): InterventionProxy like type to use as a Proxy for this Model's inputs and outputs. Can have Model specific functionality added to a new sub-class.

    Attributes:
        _model_key (str): String representing what kind of model this is. Usually hugging face repo id of model to load, path to checkpoint, or class name of custom model.
        _args (List[Any]): Positional arguments used to initialize model.
        _kwargs (Dict[str,Any]): Keyword arguments used to initialize model.
        _dispatched (bool): If the _model has been loaded yet with real parameters yet.
        _custom_model (bool): If the value passed to repoid_path_model was a custom model.
        _model (torch.nn.Module): Underlying torch module.
        _envoy (Envoy): Envoy for underlying model.
        _session (Session): Session object if in a Session.
    """

    proxy_class: Type[InterventionProxy] = InterventionProxy

    # For type hinting Envoy
    def __new__(cls, *args, **kwargs) -> Self | Envoy:
        return super().__new__(cls)

    def __init__(
        self,
        model_key: Union[str, torch.nn.Module],
        *args,
        dispatch: bool = False,
        meta_buffers: bool = True,
        **kwargs,
    ) -> None:

        self._model_key = model_key

        self._args = args
        self._kwargs = kwargs

        self._dispatched = False
        self._custom_model = False

        self._model: torch.nn.Module = None
        self._session: Session = None
        self._default_graph: Graph = None

        logger.info(f"Initializing `{self._model_key}`...")

        # Handle passing in a pre-initialized model to wrap.
        # Therefore the NNsight model is "pre-dispatched".
        if isinstance(model_key, torch.nn.Module):
            self._model_key = model_key.__class__.__name__
            self._custom_model = True
            self._dispatched = True
            self._model = model_key

        # Otherwise load from _load(...).
        if not self._custom_model:
            # Load skeleton of model by putting all tensors on meta.
            with init_empty_weights(include_buffers=meta_buffers):
                self._model = self._load(self._model_key, *args, **kwargs)

        self._envoy = Envoy(self._model)

        if dispatch and not self._dispatched:
            # Dispatch ._model on initialization vs lazy dispatching.
            self.dispatch_model()

        logger.info(f"Initialized `{self._model_key}`")

    def trace(
        self,
        *inputs: Any,
        trace: bool = True,
        invoker_args: Dict[str, Any] = None,
        backend: Union[Backend, str] = None,
        remote: bool = False,
        blocking: bool = True,
        scan: bool = False,
        **kwargs: Dict[str, Any],
    ) -> Union[Tracer, Any]:
        """Entrypoint into the tracing and interleaving functionality nnsight provides.

        In short, allows access to the future inputs and outputs of modules in order to trace what operations you would like to perform on them.
        This can be as simple as accessing and saving activations for inspection, or as complicated as transforming the activations and gradients in a forward pass over multiple inputs.

        Args:
            inputs (tuple[Any])
            trace (bool, optional): If to open a tracing context. Otherwise immediately run the model and return the raw output. Defaults to True.
            invoker_args (Dict[str, Any], optional): Keyword arguments to pass to Invoker initialization, and then downstream to the model's .prepare_inputs(...) method. Used when giving input directly to `.trace(...)`. Defaults to None.
            kwargs (Dict[str, Any]): Keyword arguments passed to Tracer initialization, and then downstream to the model's ._execute(...) method.
            backend (Union[Backend, str]): Backend for this Tracer object.
            remote (bool): Use RemoteBackend with default url.

        Raises:
            ValueError: If trace is False and no inputs were provided (nothing to run with)

        Returns:
            Union[Tracer, Any]: Either the Tracer used for tracing, or the raw output if trace is False.

        Examples:

            There are a few ways you can use ``.trace(...)`` depending in your use case.

            Lets use this extremely basic model for our examples:

            .. code-block:: python

                import torch
                from collections import OrderedDict

                input_size = 5
                hidden_dims = 10
                output_size = 2

                model = nn.Sequential(OrderedDict([
                    ('layer1', torch.nn.Linear(input_size, hidden_dims)),
                    ('sigma1', torch.nn.Sigmoid()),
                    ('layer2', torch.nn.Linear(hidden_dims, output_size)),
                    ('sigma2', torch.nn.Sigmoid()),
                ]))

                example_input = torch.rand((1, input_size))


            The first example has us running the model with a single example input, and saving the input and output of 'layer2' as well as the final output using the tracing context.

            .. code-block:: python

                from nnsight import NNsight

                with NNsight(model).trace(example_input) as model:

                    l2_input = model.layer2.input.save()
                    l2_output = model.layer2.output.save()

                    output = model.output.save()

                print(l2_input)
                print(l2_output)
                print(output)

            The second example allows us to divide up multiple inputs into one batch, and scope an inner invoker context to each one.
            We indicate this simply by not passing and positional inputs into `.trace(...)`. The Tracer object then expects you to enter each input via `Tracer.invoke(...)`

            .. code-block:: python

                example_input2 = torch.rand((1, input_size))

                with NNsight(model).trace() as model:

                    with model.invoke(example_input):

                        output1 = model.output.save()

                    with model.invoke(example_input2):

                        output2 = model.output.save()

                print(output1)
                print(output2)
        """

        # TODO raise error/warning if trying to use one backend with another condition satisfied?

        bridge = None

        if backend is not None:
            pass

        elif self._session is not None:

            backend = BridgeBackend(weakref.proxy(self._session.bridge))

            bridge = self._session.bridge

        # If remote, use RemoteBackend with default url.
        elif remote:

            backend = RemoteBackend(blocking=blocking)

        # By default, use LocalBackend.
        elif backend is None:

            backend = LocalBackend()

        # If backend is a string, assume RemoteBackend url.
        elif isinstance(backend, str):

            backend = RemoteBackend(host=backend, blocking=blocking)

        # Create Tracer object.
        if self._default_graph is not None:

            graph = self._default_graph.copy()

            tracer = Tracer(backend, self, bridge=bridge, graph=graph, **kwargs)
        else:
            tracer = Tracer(backend, self, bridge=bridge, **kwargs)

        # If user provided input directly to .trace(...).
        if len(inputs) > 0:

            invoker_args = invoker_args or {}

            invoker_args["scan"] = scan

            # If trace is False, we'll enter the Tracer context immediately and enter an Invoker context with the provided inputs as well.
            # We'll also save the output of the model and return its value directly.
            if not trace:

                with tracer:
                    with tracer.invoke(*inputs, **invoker_args):

                        output = self._envoy.output.save()

                return output.value

            # Otherwise open an invoker context with the give args.
            tracer.invoke(*inputs, **invoker_args)

        # If trace is False, you had to have provided an input.
        if not trace:

            raise ValueError("Can't execute on no inputs!")

        return tracer

    def scan(self, *inputs, **kwargs) -> Tracer:
        """Context just to populate fake tenor proxy values using scan and validate.
        Useful when looking for just the shapes of future tensors

        Examples:

            .. code-block:: python

                with model.scan(" "):

                    dim = model.module.output.shape[-1]

                print(dim)

        Returns:
            Tracer: Tracer context with Noop backend.
        """

        return self.trace(
            *inputs, **kwargs, scan=True, validate=True, backend=NoopBackend()
        )

    def edit(
        self,
        *inputs: Any,
        inplace: bool = False,
        return_context: bool = False,
        **kwargs: Dict[str, Any],
    ) -> Union[Tracer, Any]:
        """Create a trace context with an edit backend and apply a list of edits.

        The edit backend sets a default graph on an NNsight model copy which is
        run on future trace calls.

        This operation is not inplace!

        Args:
            inputs (tuple[Any])
            inplace (bool): If True, makes edits in-place.
            return_context (bool): If True, returns the editor Tracer context.
            kwargs (Dict[str, Any]): Keyword arguments passed to Tracer initialization, and then downstream to the model's ._execute(...) method.

        Returns:
            Union[Tracer, Any]: Either the Tracer used for tracing, or the raw output if trace is False.

        Example:
            .. code-block:: python
            from nnsight import LanguageModel

            gpt2 = LanguageModel("openai-community/gpt2)

            class ComplexModule(torch.nn.Module):
                def __init__(self):
                    super().__init__()
                    self.one = WrapperModule()

                def forward(self, x):
                    return self.one(x)

            l0 = gpt2.transformer.h[0]
            l0.attachment = ComplexModule()

            with gpt2.edit("test") as gpt2_edited:
                acts = l0.output[0]
                l0.output[0][:] = l0.attachment(acts, hook=True)

            with gpt2.trace(MSG_prompt):
                original = l0.output[0].clone().save()
                l0.output[0][:] *= 0.0
                original_output = gpt2.output.logits.save()

            with gpt2_edited.trace(MSG_prompt):
                one = l0.attachment.one.output.clone().save()
                l0.attachment.output *= 0.0
                edited_output = gpt2.output.logits.save()

            print(original_output)
            print(edited_output)
        """
        model_to_edit = self
        if not inplace:
            model_to_edit = self._shallow_copy()

        return model_to_edit.trace(
            *inputs,
            validate=kwargs.pop("validate", False),
            return_context=return_context,
            **kwargs,
            backend=EditBackend(),
        )

    def session(
        self,
        backend: Union[Backend, str] = None,
        remote: bool = False,
        blocking: bool = True,
        **kwargs,
    ) -> Session:
        """Create a session context using a Session.

        Args:
            backend (Backend): Backend for this Session object.
            remote (bool): Use RemoteBackend with default url.

        Returns:
            Session: Session.
        """

        # If remote, use RemoteBackend with default url.
        if remote:

            backend = RemoteBackend(blocking=blocking)

        # By default, use LocalBackend.
        elif backend is None:

            backend = LocalBackend()

        # If backend is a string, assume RemoteBackend url.
        elif isinstance(backend, str):

            backend = RemoteBackend(host=backend, blocking=blocking)

        session = Session(backend, self, **kwargs)

        self._session = session

        return session

    def interleave(
        self,
        fn: Callable,
        intervention_graph: Graph,
        *inputs: List[List[Any]],
        **kwargs,
    ) -> None:
        """Runs some function with some inputs and some graph with the appropriate contexts for this model.

        Loads and dispatched ._model if not already done so.

        Re-compiles Graph with ._model to prepare for a new execution of graph.

        Runs ._prepare_inputs(...) one last time to get total_batch_size.

        Handles adding and removing hooks on modules via HookHandler and tracking number of times a module has been called via InterventionHandler.

        After execution, garbage collects and clears cuda memory.

        Args:
            fn (Callable): Function or method to run.
            intervention_graph (Graph): Intervention graph to interleave with model's computation graph.
            inputs (List[List[Any]]): List of multiple groups of inputs to give to function for each invoker.
        """

        # Loads and dispatched ._model if not already done so.
        if not self._dispatched:
            self.dispatch_model()

        logger.info(f"Running `{self._model_key}`...")

        # We need to pre-process and batch all inputs as (inputs) is a list of each set of inputs from each invocation.
        batch_groups = []
        batch_start = 0
        batched_input = None
        batch_size = 0

        for _inputs in inputs:

            _inputs, batch_size = self._prepare_inputs(*_inputs)

            batch_groups.append((batch_start, batch_size))
            batch_start += batch_size

            batched_input = self._batch_inputs(batched_input, *_inputs)

        if len(inputs) > 0:
            inputs, batch_size = self._prepare_inputs(*batched_input)

        intervention_handler = InterventionHandler(
            intervention_graph, batch_groups, batch_size
        )

        module_paths = InterventionProtocol.get_interventions(intervention_graph).keys()

        with HookHandler(
            self._model,
            list(module_paths),
            input_hook=lambda activations, module_path: InterventionProtocol.intervene(
                activations, module_path, "input", intervention_handler
            ),
            output_hook=lambda activations, module_path: InterventionProtocol.intervene(
                activations, module_path, "output", intervention_handler
            ),
        ):
            try:
                fn(*inputs, **kwargs)
            except protocols.EarlyStopProtocol.EarlyStopException:
                # TODO: Log.
                for node in intervention_graph.nodes.values():
                    if not node.executed():
                        node.clean()

        logger.info(f"Completed `{self._model_key}`")

    def dispatch_model(self, *args, **kwargs) -> None:
        """Dispatch ._model to have real parameters  using ._load(...)."""

        logger.info(f"Dispatching `{self._model_key}`...")

        self._model = self._load(
            self._model_key, *self._args, *args, **kwargs, **self._kwargs
        )

        self._envoy._update(self._model)

        self._dispatched = True

        logger.info(f"Dispatched `{self._model_key}`")

    def to(self, *args, **kwargs) -> Self:
        """Override torch.nn.Module.to so this returns the NNSight model, not the underlying module when doing: model = model.to(...)

        Returns:
            Envoy: Envoy.
        """

        self._model = self._model.to(*args, **kwargs)

        return self

    def clear_edits(self) -> None:
        """Resets the default graph of this model."""
        self._default_graph = None

    def __repr__(self) -> str:
        """Wrapper of ._model's representation as the NNsight model's representation.

        Returns:
            str: Representation.
        """
        return repr(self._envoy)

    def __setattr__(self, key: Any, value: Any) -> None:
        """Overload setattr to create and set an Envoy when trying to set a torch Module."""

        if key not in ("_model", "_model_key") and isinstance(value, torch.nn.Module):

            setattr(self._envoy, key, value)

        else:

            object.__setattr__(self, key, value)

    def __getattr__(self, key: Any) -> Union[Envoy, InterventionProxy, Any]:
        """Wrapper of ._envoy's attributes to access module's inputs and outputs.

        Returns:
            Any: Attribute.
        """
        return getattr(self._envoy, key)

    ### NNsight VIRTUAL METHODS BELOW #####################################

    def _load(self, repo_id: str, *args, **kwargs) -> torch.nn.Module:
        """Virtual method to load the model from scratch.

        Default implementation loads a model from AutoModel.from_config if not dispatched, else uses accelerate.load_checkpoint_and_dispatch.

        Args:
            model_key (str): String value used to load model. Usually huggingface repo_id or checkpoint path.

        Returns:
            torch.nn.Module: Model.
        """

        if self._model is None:

            config = AutoConfig.from_pretrained(repo_id, *args, **kwargs)

            return AutoModel.from_config(config, trust_remote_code=True)

        return accelerate.load_checkpoint_and_dispatch(self._model, repo_id, **kwargs)

    def _execute(self, *prepared_inputs: Any, **kwargs) -> Any:
        """Virtual method to run the underlying ._model with some inputs.

        Default implementation util.applies moving all tensors to the device of the first parameter in ._model and passes the values into the model.

        Args:
            prepared_inputs (tuple[Any]): Prepared inputs.
        """

        try:
            device = next(self._model.parameters()).device

            prepared_inputs = util.apply(
                prepared_inputs, lambda x: x.to(device), torch.Tensor
            )

        except:
            pass

        return self._model(
            *prepared_inputs,
            **kwargs,
        )

    def _prepare_inputs(self, *inputs: Any, **kwargs) -> Tuple[Tuple[Any], int]:
        """Virtual method to prepare inputs before batching and execution and return batch size of prepared_inputs.

        Default implementation just returns inputs and length of first input.

        Args:
            inputs (tuple[Any]): Inputs to prepare for batching and execution.
            int: Batch size of prepared_inputs.

        Returns:
            Tuple[tuple[Any], int]: Prepared inputs, batch size of inputs.
        """
        return inputs, len(inputs[0])

    def _batch_inputs(
        self,
        batched_inputs: Optional[Any],
        *prepared_inputs: Any,
    ) -> Any:
        """Virtual method to batch together results from _prepare_inputs.

        Default implementation returns list of all prepared_inputs.

        Args:
            batched_inputs (Any): Current state of batched_inputs. Initially None.
            prepared_inputs (tuple[Any]): Most recent result from _prepare_inputs.

        Returns:
            Any: Batched inputs.
        """

        if batched_inputs is None:
            batched_inputs = prepared_inputs

        else:

            batched_inputs = tuple(
                [
                    torch.concatenate((batched_inputs[i], prepared_inputs[i]))
                    for i in range(len(batched_inputs))
                ]
            )

        return batched_inputs

    def _shallow_copy(self) -> Self:
        """Creates a new instance copy of the same class with the all the attributes of the original instance.

        Returns:
            Self: NNsightModel
        """
        copy = self.__class__.__new__(self.__class__)
        for key, value in self.__dict__.items():
            copy.__dict__[key] = value

        return copy



"""Module for utility functions and classes used throughout the package."""

import importlib
import types
from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Collection,
    Dict,
    Optional,
    Tuple,
    Type,
    Union,
)

import torch

if TYPE_CHECKING:
    from .tracing.Node import Node

# TODO Have an Exception you can raise to stop apply early

def apply(
    data: Any, fn: Callable, cls: Type, inplace: bool = False
) -> Collection:
    """Applies some function to all members of a collection of a give type (or types)

    Args:
        data (Any): Collection of data to apply function to.
        fn (Callable): Function to apply.
        cls (type): Type or Types to apply function to.
        inplace (bool): If to apply the fn inplace. (For lists and dicts)

    Returns:
        Any: Same kind of collection as data, after then fn has been applied to members of given type.
    """

    if isinstance(data, cls):
        return fn(data)

    data_type = type(data)

    if data_type == list:
        if inplace:
            for idx, _data in enumerate(data):
                data[idx] = apply(_data, fn, cls, inplace=inplace)
            return data
        return [apply(_data, fn, cls, inplace=inplace) for _data in data]

    elif data_type == tuple:
        return tuple([apply(_data, fn, cls, inplace=inplace) for _data in data])

    elif data_type == dict:
        if inplace:
            for key, value in data.items():
                data[key] = apply(value, fn, cls, inplace=inplace)
            return data
        return {
            key: apply(value, fn, cls, inplace=inplace)
            for key, value in data.items()
        }

    elif data_type == slice:
        return slice(
            apply(data.start, fn, cls, inplace=inplace),
            apply(data.stop, fn, cls, inplace=inplace),
            apply(data.step, fn, cls, inplace=inplace),
        )

    return data


def fetch_attr(object: object, target: str) -> Any:
    """Retrieves an attribute from an object hierarchy given an attribute path. Levels are separated by '.' e.x (transformer.h.1)

    Args:
        object (object): Root object to get attribute from.
        target (str): Attribute path as '.' separated string.

    Returns:
        Any: Fetched attribute.
    """
    if target == "":
        return object

    target_atoms = target.split(".")

    for atom in target_atoms:

        if not atom:
            continue

        object = getattr(object, atom)

    return object


def wrap(object: object, wrapper: Type, *args, **kwargs) -> object:
    """Wraps some object given some wrapper type.
    Updates the __class__ attribute of the object and calls the wrapper type's __init__ method.

    Args:
        object (object): Object to wrap.
        wrapper (Type): Type to wrap the object in.

    Returns:
        object: Wrapped object.
    """
    if isinstance(object, wrapper):
        return object

    new_class = types.new_class(
        object.__class__.__name__,
        (object.__class__, wrapper),
    )

    object.__class__ = new_class

    wrapper.__init__(object, *args, **kwargs)

    return object


def to_import_path(type: type) -> str:

    return f"{type.__module__}.{type.__name__}"


def from_import_path(import_path: str) -> type:

    *import_path, classname = import_path.split(".")
    import_path = ".".join(import_path)

    return getattr(importlib.import_module(import_path), classname)


class WrapperModule(torch.nn.Module):
    """Simple torch module which passes it's input through. Useful for hooking.
    If there is only one argument, returns the first element.
    """

    def forward(self, *args, **kwargs):
        if len(args) == 1:
            args = args[0]

        return args



import pytest
import torch

import nnsight
from nnsight.contexts.GraphBasedContext import GlobalTracingContext
from nnsight.contexts.Tracer import Tracer
from nnsight.schema.Request import RequestModel
from nnsight.tracing.Graph import Graph


@pytest.fixture(scope="module")
def gpt2(device: str):
    return nnsight.LanguageModel(
        "openai-community/gpt2", device_map=device, dispatch=True
    )


@pytest.fixture
def MSG_prompt():
    return "Madison Square Garden is located in the city of"


def _test_serialize(tracer: Tracer):
    with GlobalTracingContext.exit_global_tracing_context():
        request = RequestModel(
            object=tracer, model_key=tracer.remote_backend_get_model_key()
        )
        request_json = request.model_dump(
            mode="json", exclude=["session_id", "received", "id"]
        )

        request2 = RequestModel(**request_json)
        tracer = request2.deserialize(tracer.model)
    assert isinstance(tracer.graph, Graph)


@torch.no_grad()
def test_generation(gpt2: nnsight.LanguageModel, MSG_prompt: str):
    with gpt2.generate(max_new_tokens=3, validate=True) as generator:
        with generator.invoke(MSG_prompt, scan=True) as invoker:
            output = gpt2.generator.output.save()

        _test_serialize(generator)

    output = gpt2.tokenizer.decode(output.value[0])

    assert (
        output
        == "Madison Square Garden is located in the city of New York City"
    )


@torch.no_grad()
def test_save(gpt2: nnsight.LanguageModel):
    with gpt2.generate("Hello world", validate=True, scan=True) as tracer:

        hs = gpt2.transformer.h[-1].output[0].save()
        hs_input = gpt2.transformer.h[-1].input.save()

        _test_serialize(tracer)

    assert hs.value is not None
    assert isinstance(hs.value, torch.Tensor)
    assert hs.value.ndim == 3

    assert hs_input.value is not None
    assert isinstance(hs_input.value, torch.Tensor)
    assert hs_input.value.ndim == 3


@torch.no_grad()
def test_set1(gpt2: nnsight.LanguageModel, MSG_prompt: str):
    with gpt2.generate(validate=True) as tracer:
        with tracer.invoke(MSG_prompt, scan=True) as invoker:
            pre = gpt2.transformer.h[-1].output[0].clone().save()

            gpt2.transformer.h[-1].output[0][:] = 0

            post = gpt2.transformer.h[-1].output[0].save()

            output = gpt2.generator.output.save()

        _test_serialize(tracer)

    output = gpt2.tokenizer.decode(output.value[0])

    assert not (pre.value == 0).all().item()
    assert (post.value == 0).all().item()
    assert output != "Madison Square Garden is located in the city of New"


@torch.no_grad()
def test_set2(gpt2: nnsight.LanguageModel, MSG_prompt: str):
    with gpt2.generate(validate=True) as generator:
        with generator.invoke(MSG_prompt, scan=True) as invoker:
            pre = gpt2.transformer.wte.output.clone().save()

            gpt2.transformer.wte.output = gpt2.transformer.wte.output * 0

            post = gpt2.transformer.wte.output.save()

            output = gpt2.generator.output.save()

        _test_serialize(generator)

    output = gpt2.tokenizer.decode(output.value[0])

    assert not (pre.value == 0).all().item()
    assert (post.value == 0).all().item()
    assert output != "Madison Square Garden is located in the city of New"


@torch.no_grad()
def test_adhoc_module(gpt2: nnsight.LanguageModel):
    with gpt2.generate(validate=True) as generator:
        with generator.invoke(
            "The Eiffel Tower is in the city of", scan=True
        ) as invoker:
            hidden_states = gpt2.transformer.h[-1].output[0]
            hidden_states = gpt2.lm_head(gpt2.transformer.ln_f(hidden_states))
            tokens = torch.softmax(hidden_states, dim=2).argmax(dim=2).save()

        _test_serialize(generator)

    output = gpt2.tokenizer.decode(tokens.value[0])

    assert output == "\n-el Tower is a the middle centre Paris"


@torch.no_grad()
def test_embeddings_set1(gpt2: nnsight.LanguageModel, MSG_prompt: str):
    with gpt2.generate(max_new_tokens=3, validate=True) as generator:
        with generator.invoke(MSG_prompt, scan=True) as invoker:
            embeddings = gpt2.transformer.wte.output

            output1 = gpt2.generator.output.save()

        with generator.invoke("_ _ _ _ _ _ _ _ _", scan=True) as invoker:
            gpt2.transformer.wte.output = embeddings

            output2 = gpt2.generator.output.save()

        _test_serialize(generator)

    output1 = gpt2.tokenizer.decode(output1.value[0])
    output2 = gpt2.tokenizer.decode(output2.value[0])

    assert (
        output1
        == "Madison Square Garden is located in the city of New York City"
    )
    assert output2 == "_ _ _ _ _ _ _ _ _ New York City"


@torch.no_grad()
def test_embeddings_set2(gpt2: nnsight.LanguageModel, MSG_prompt: str):
    with gpt2.generate(max_new_tokens=3, validate=True) as generator:
        with generator.invoke(MSG_prompt, scan=True) as invoker:
            embeddings = gpt2.transformer.wte.output.save()

            output = gpt2.generator.output.save()

    output1 = gpt2.tokenizer.decode(output.value[0])

    with gpt2.generate(max_new_tokens=3, validate=True) as generator:
        with generator.invoke("_ _ _ _ _ _ _ _ _", scan=True) as invoker:
            gpt2.transformer.wte.output = embeddings.value

            output = gpt2.generator.output.save()

        _test_serialize(generator)

    output2 = gpt2.tokenizer.decode(output.value[0])

    assert (
        output1
        == "Madison Square Garden is located in the city of New York City"
    )
    assert output2 == "_ _ _ _ _ _ _ _ _ New York City"


def test_retain_grad(gpt2: nnsight.LanguageModel):
    with gpt2.trace(validate=True) as tracer:
        with tracer.invoke("Hello World", scan=True) as invoker:
            hidden_states = gpt2.transformer.h[-1].output[0].save()
            hidden_states.retain_grad()

            logits = gpt2.lm_head.output

            logits.sum().backward()

        _test_serialize(tracer)

    assert hidden_states.value.grad is not None


def test_grad(gpt2: nnsight.LanguageModel):
    with gpt2.trace(validate=True) as tracer:
        with tracer.invoke("Hello World", scan=True) as invoker:
            hidden_states = gpt2.transformer.h[-1].output[0].save()
            hidden_states_grad = hidden_states.grad.save()
            hidden_states_grad[:] = 0

            logits = gpt2.lm_head.output

            logits.sum().backward()

        _test_serialize(tracer)

    hidden_states.value

    assert (hidden_states_grad.value == 0).all().item()

    with gpt2.trace(validate=True) as tracer:
        with tracer.invoke("Hello World", scan=True) as invoker:
            hidden_states = gpt2.transformer.h[-1].output[0].save()
            grad = hidden_states.grad.clone()
            grad[:] = 0
            hidden_states.grad = grad

            logits = gpt2.lm_head.output

            logits.sum().backward()

        _test_serialize(tracer)

    hidden_states.value
    assert (hidden_states_grad.value == 0).all().item()


def test_other_device_tensors(gpt2: nnsight.LanguageModel):

    device = next(gpt2.parameters())

    lin = torch.nn.Linear(768, 768).to(device)
    bias = torch.randn(768).to(device)

    def fun(x):
        return torch.nn.ReLU()(lin(x) - bias)

    with gpt2.trace("fish", validate=True, scan=True) as tracer:
        x = gpt2.transformer.h[0].mlp.output
        y = fun(x)
        z = y.save()

        # TODO
        # _test_serialize(tracer)

    z.value


def test_multi_grad(gpt2: nnsight.LanguageModel):
    with gpt2.trace(validate=True) as tracer:
        with tracer.invoke("Hello World", scan=True) as invoker:
            hidden_states = gpt2.transformer.h[-1].output[0].save()

            hidden_states_grad1 = hidden_states.grad.save()

            logits = gpt2.lm_head.output

            logits.sum().backward(retain_graph=True)

            hidden_states_grad2 = hidden_states.grad.save()

            logits = logits * 2

            logits.sum().backward()

        _test_serialize(tracer)

    assert not torch.all(hidden_states_grad1.eq(hidden_states_grad2))


def test_editing(gpt2: nnsight.LanguageModel, MSG_prompt: str):
    from nnsight.util import WrapperModule

    class ComplexModule(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.one = WrapperModule()

        def forward(self, x):
            return self.one(x)

    l0 = gpt2.transformer.h[0]
    l0.attachment = ComplexModule()

    with gpt2.edit() as gpt2_edited:
        acts = l0.output[0]
        l0.output[0][:] = l0.attachment(acts, hook=True)

    # Get values pre editing
    with gpt2.trace(MSG_prompt):
        original = l0.output[0].clone().save()
        l0.output[0][:] *= 0.0
        original_output = gpt2.output.logits.save()

    with gpt2_edited.trace(MSG_prompt):
        one = l0.attachment.one.output.clone().save()
        l0.attachment.output *= 0.0
        edited_output = gpt2.output.logits.save()

    # Check that submodule in attached model
    # is equal to original output.
    assert torch.equal(original, one)
    # Check that edits propagate from attached module
    assert torch.equal(original_output, edited_output)


def test_non_inplace_editing(gpt2: nnsight.LanguageModel, MSG_prompt: str):

    with gpt2.edit(inplace=True):
        gpt2.transformer.h[1].output[0][:, 0] = 0

    with gpt2.edit() as gpt2_edited:
        gpt2.transformer.h[1].output[0][:, 1] = 0

    with gpt2.trace(MSG_prompt):
        l1_out = gpt2.transformer.h[1].output[0].save()

    with gpt2_edited.trace(MSG_prompt):
        l1_out_edited = gpt2_edited.transformer.h[1].output[0].save()

    assert torch.all(l1_out[:, 0] == 0) and torch.all(l1_out[:, 1] != 0)
    assert torch.all(l1_out_edited[:, 0] == 0) and torch.all(l1_out_edited[: , 1] == 0)


def test_clear_edits(gpt2: nnsight.LanguageModel, MSG_prompt: str):
    with gpt2.edit(inplace=True):
        gpt2.transformer.h[1].output[0][:] = 0

    with gpt2.trace(MSG_prompt):
        l1_out = gpt2.transformer.h[1].output[0].save()

    gpt2.clear_edits()

    with gpt2.trace(MSG_prompt):
        l1_out_unedited = gpt2.transformer.h[1].output[0].save()

    assert torch.all(l1_out == 0)
    assert torch.all(l1_out_unedited != 0)


def test_batched_editing(gpt2: nnsight.LanguageModel):
    from nnsight.util import WrapperModule

    class ComplexModule(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.one = WrapperModule()

        def forward(self, x):
            return self.one(x)

    l0 = gpt2.transformer.h[0]
    l0.attachment = ComplexModule()

    batch = ["a", "b"]
    single = "a"

    with gpt2.edit(single) as gpt2_edited:
        acts = l0.output[0]
        l0.output[0][:] = l0.attachment(acts, hook=True)

    with gpt2_edited.trace(batch):
        edited = l0.attachment.output.save()

    # Check that the batch size does not narrow
    assert edited.shape[0] == 2


def test_conditional_interventions(gpt2: nnsight.LanguageModel):
    with gpt2.session() as session:
        with gpt2.trace("Hello World", validate=True, scan=True) as tracer:
            with tracer.cond(
                torch.all(gpt2.transformer.h[5].output[0] < 100000)
            ):
                gpt2.transformer.h[-1].output[0][:] = 0

            output = gpt2.transformer.h[-1].output[0].save()

        _test_serialize(session)

    assert torch.all(output.value == 0)


def test_input_setting(gpt2: nnsight.LanguageModel, MSG_prompt: str):
    with gpt2.session():
        with gpt2.trace(MSG_prompt):
            hs = gpt2.transformer.h[6].inputs
            tokens_out_1 = gpt2.lm_head.output.argmax(dim=-1).save()

        with gpt2.trace(MSG_prompt):
            gpt2.transformer.h[6].input = hs[0][0]
            tokens_out_2 = gpt2.lm_head.output.argmax(dim=-1).save()

    prediction_1 = gpt2.tokenizer.decode(tokens_out_1[0][-1])
    prediction_2 = gpt2.tokenizer.decode(tokens_out_2[0][-1])

    assert prediction_1 == prediction_2



from __future__ import annotations

import weakref
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union, Tuple

from typing_extensions import Self

from ..tracing import protocols
from ..tracing.Bridge import Bridge
from ..tracing.Graph import Graph
from . import resolve_dependencies
from .backends import Backend, EditBackend, BridgeMixin, EditMixin, RemoteMixin
from .GraphBasedContext import GraphBasedContext
from .Invoker import Invoker

if TYPE_CHECKING:
    from ..models.mixins import RemoteableMixin
    from ..models.NNsightModel import NNsight


class Tracer(GraphBasedContext, RemoteMixin, BridgeMixin, EditMixin):
    """The Tracer class creates a :class:`nnsight.tracing.Graph.Graph` around the ._model of a :class:`nnsight.models.NNsightModel.NNsight` which tracks and manages the operations performed on the inputs and outputs of said model.

    Attributes:
        _model (nnsight.models.NNsightModel.NNsight): nnsight Model object that ths context manager traces and executes.
        _graph (nnsight.tracing.Graph.Graph): Graph which traces operations performed on the input and output of modules' Envoys are added and later executed.
        _args (List[Any]): Positional arguments to be passed to function that executes the model.
        _kwargs (Dict[str,Any]): Keyword arguments to be passed to function that executes the model.
        _invoker_inputs (List[Any]): Inputs for each invocation of this Tracer.
        _invoker (Invoker): Currently open Invoker.
    """

    def __init__(
        self,
        backend: Backend,
        model: "NNsight",
        validate: bool = False,
        graph: Graph = None,
        bridge: Bridge = None,
        return_context: bool = False,
        **kwargs,
    ) -> None:

        self.model = model

        self.return_context = return_context

        GraphBasedContext.__init__(
            self,
            backend,
            graph=graph,
            bridge=bridge,
            proxy_class=model.proxy_class,
            validate=validate,
            sequential=False,
        )

        protocols.ApplyModuleProtocol.set_module(self.graph, self.model)

        self._kwargs = kwargs

        self.invoker: Optional[Invoker] = None

        self._invoker_inputs: List[Any] = []

        # Module Envoys need to know about the current Tracer to create the correct proxies.
        self.model._envoy._set_tracer(weakref.proxy(self))

    def __getattr__(self, key: Any) -> Any:
        """Wrapper of .model._envoy's attributes to access module Envoy inputs and outputs.

        Returns:
            Any: Attribute.
        """
        return getattr(self.model._envoy, key)

    def __enter__(self) -> Union[Self, "NNsight", Tuple["NNsight", Self]]:

        tracer = super().__enter__()

        if self.invoker is not None:

            self.invoker.__enter__()

        if isinstance(self.backend, EditBackend):
            if self.return_context:
                return self.model, self
            
            return self.model

        return tracer

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:

        if self.invoker is not None:

            self.invoker.__exit__(None, None, None)

        self.model._envoy._reset()


        super().__exit__(exc_type, exc_val, exc_tb)

    def invoke(self, *inputs: Any, **kwargs) -> Invoker:
        """Create an Invoker context dor a given input.

        Raises:
            Exception: If an Invoker context is already open

        Returns:
            Invoker: Invoker.
        """

        if self.invoker is not None:

            raise Exception("Can't create an invoker context with one already open!")

        return Invoker(self, *inputs, **kwargs)

    def next(self, increment: int = 1) -> None:
        """Increments call_iter of all module Envoys. Useful when doing iterative/generative runs.

        Args:
            increment (int): How many call_iter to increment at once. Defaults to 1.
        """

        self.model._envoy.next(increment=increment, propagate=True)

    ##### BACKENDS ###############################

    def local_backend_execute(self) -> Graph:

        protocols.ApplyModuleProtocol.set_module(self.graph, self.model._model)

        self.graph.reset()

        invoker_inputs = self._invoker_inputs

        # If ths graph has a Bridge, we need to check for Nodes in the input itself.
        if protocols.BridgeProtocol.has_bridge(self.graph):

            invoker_inputs = resolve_dependencies(invoker_inputs)

        self.graph.execute()

        self.model.interleave(
            self.model._execute,
            self.graph,
            *invoker_inputs,
            **self._kwargs,
        )

        graph = self.graph
        graph.alive = False

        if not isinstance(graph, weakref.ProxyType):
            self.graph = weakref.proxy(graph)

        return graph

    def edit_backend_execute(self) -> Graph:

        self.model._default_graph = self.graph

    def remote_backend_get_model_key(self) -> str:

        self.model: "RemoteableMixin"

        return self.model.to_model_key()

    def remote_backend_postprocess_result(self, local_result: Graph) -> Dict[str, Any]:

        from ..schema.Response import ResultModel

        return ResultModel.from_graph(local_result)

    def remote_backend_handle_result_value(self, value: Dict[str, Any]) -> None:

        # TODO : graph mismatch handle. hash json ?
        for node_name, node_value in value.items():
            self.graph.nodes[node_name]._value = node_value

    def remote_backend_cleanup(self):
        
        graph = self.graph
        graph.alive = False

        if not isinstance(graph, weakref.ProxyType):
            self.graph = weakref.proxy(graph)

    def __repr__(self) -> str:
        return f"&lt;{self.__class__.__name__} at {hex(id(self))}&gt;"



from typing import Any


class Backend:
    """A backend is what executes a context object when it __exit__s."""

    def __call__(self, obj: Any) -> None:
        """Handles execution of a context object on exit. (like a Tracer or Session).

        Args:
            obj (Any): Context object to execute.
        """

        raise NotImplementedError()


from .BridgeBackend import BridgeBackend, BridgeMixin
from .EditBackend import EditBackend, EditMixin
from .LocalBackend import LocalBackend, LocalMixin
from .NoopBackend import NoopBackend
from .RemoteBackend import RemoteBackend, RemoteMixin



def pytest_addoption(parser):
    parser.addoption("--device", action="store", default="cuda:0")


def pytest_generate_tests(metafunc):
    # This is called for every test. Only get/set command line arguments
    # if the argument is specified in the list of test "fixturenames".
    option_value = metafunc.config.option.device
    if "device" in metafunc.fixturenames and option_value is not None:
        metafunc.parametrize("device", [option_value], scope="module")


collect_ignore = ["examples/test_server.py", "examples/test_server_llama.py"]
